"""
TODO: this docstring is the main xsar module documentation shown to the user. It's should be updated with some examples.
"""

import warnings

from importlib import metadata

__version__ = metadata.version("xsar")

import logging
from .utils import timing, config, url_get
import os
import zipfile
import pandas as pd
import geopandas as gpd


from .sentinel1_meta import Sentinel1Meta
from .sentinel1_dataset import Sentinel1Dataset
from .radarsat2_meta import RadarSat2Meta
from .radarsat2_dataset import RadarSat2Dataset
from .rcm_meta import RcmMeta
from .rcm_dataset import RcmDataset

logger = logging.getLogger("xsar")
logger.addHandler(logging.NullHandler())

os.environ["GDAL_CACHEMAX"] = "128"


@timing
def open_dataset(*args, **kwargs):
    """
    Parameters
    ----------
    *args:
        Passed to `xsar.SentinelDataset`
    **kwargs:
        Passed to `xsar.SentinelDataset`

    Returns
    -------
    xarray.Dataset

    Notes
    -----
    xsar.open_dataset` is a simple wrapper to `xsar.SentinelDataset` that directly returns the `xarray.Dataset` object.

    >>> xsar.Sentinel1Dataset(*args, **kwargs).dataset
    >>> xsar.RadarSat2Dataset(*args, **kwargs).dataset
    >>> xsar.RcmDataset(*args, **kwargs).dataset

    See Also
    --------
    xsar.Sentinel1Dataset
    xsar.RadarSat2Dataset
    xsar.RcmDataset
    """
    dataset_id = args[0]
    # TODO: check product type (S1, RS2), and call specific reader
    if (
        isinstance(dataset_id, Sentinel1Meta)
        or isinstance(dataset_id, str)
        and "S1" in dataset_id
    ):
        sar_obj = Sentinel1Dataset(*args, **kwargs)
    elif (
        isinstance(dataset_id, RadarSat2Meta)
        or isinstance(dataset_id, str)
        and "RS2" in dataset_id
    ):
        sar_obj = RadarSat2Dataset(*args, **kwargs)
    elif (
        isinstance(dataset_id, RcmMeta)
        or isinstance(dataset_id, str)
        and "RCM" in dataset_id
    ):
        sar_obj = RcmDataset(*args, **kwargs)
    else:
        raise TypeError("Unknown dataset type from %s" % str(dataset_id))
    ds = sar_obj.dataset
    return ds


def open_datatree(*args, **kwargs):
    """
    Parameters
    ----------
    *args:
        Passed to `xsar.SentinelDataset`
    **kwargs:
        Passed to `xsar.SentinelDataset`

    Returns
    -------
    xarray.Dataset

    Notes
    -----
    xsar.open_dataset` is a simple wrapper to `xsar.SentinelDataset` that directly returns the `xarray.Dataset` object.

    >>> xsar.Sentinel1Dataset(*args, **kwargs).dataset
    >>> xsar.RadarSat2Dataset(*args, **kwargs).dataset
    >>> xsar.RcmDataset(*args, **kwargs).dataset

    See Also
    --------
    xsar.Sentinel1Dataset
    xsar.RadarSat2Dataset
    xsar.RcmDataset
    """
    dataset_id = args[0]
    # Check product type (S1, RS2), and call specific reader
    if (
        isinstance(dataset_id, Sentinel1Meta)
        or isinstance(dataset_id, str)
        and "S1" in dataset_id
    ):
        sar_obj = Sentinel1Dataset(*args, **kwargs)
    elif (
        isinstance(dataset_id, RadarSat2Meta)
        or isinstance(dataset_id, str)
        and "RS2" in dataset_id
    ):
        sar_obj = RadarSat2Dataset(*args, **kwargs)
    elif (
        isinstance(dataset_id, RcmMeta)
        or isinstance(dataset_id, str)
        and "RCM" in dataset_id
    ):
        sar_obj = RcmDataset(*args, **kwargs)
    else:
        raise TypeError("Unknown dataset type from %s" % str(dataset_id))
    dt = sar_obj.datatree
    return dt


def product_info(path, columns="minimal", include_multi=False):
    """

    Parameters
    ----------
    path: str or iterable of str
        path or gdal url.
    columns: list of str or str, optional
        'minimal' by default: only include columns from attributes found in manifest.safe.
        Use 'spatial' to have 'time_range' and 'geometry'.
        Might be a list of properties from `xsar.Sentinel1Meta`
    include_multi: bool, optional
        False by default: don't include multi datasets

    Returns
    -------
    geopandas.GeoDataFrame
      One dataset per lines, with info as columns

    See Also
    --------
    xsar.Sentinel1Meta

    """

    info_keys = {
        "minimal": ["name", "ipf", "platform", "swath", "product", "pols", "meta"]
    }
    info_keys["spatial"] = info_keys["minimal"] + ["time_range", "geometry"]

    if isinstance(columns, str):
        columns = info_keys[columns]

    # 'meta' column is not a Sentinel1Meta attribute
    real_cols = [c for c in columns if c != "meta"]
    add_cols = []
    if "path" not in real_cols:
        add_cols.append("path")
    if "dsid" not in real_cols:
        add_cols.append("dsid")

    def _meta2df(meta):
        df = pd.Series(data=meta.to_dict(add_cols + real_cols)).to_frame().T
        if "meta" in columns:
            df["meta"] = meta
        return df

    if isinstance(path, str):
        path = [path]

    df_list = []
    for p in path:
        s1meta = Sentinel1Meta(p)
        if s1meta.multidataset and include_multi:
            df_list.append(_meta2df(s1meta))
        elif not s1meta.multidataset:
            df_list.append(_meta2df(s1meta))
        if s1meta.multidataset:
            for n in s1meta.subdatasets.index:
                s1meta = Sentinel1Meta(n)
                df_list.append(_meta2df(s1meta))
    df = pd.concat(df_list).reset_index(drop=True)
    if "geometry" in df:
        df = gpd.GeoDataFrame(df).set_crs(epsg=4326)

    df = df.set_index(["path", "dsid"], drop=False)
    if add_cols:
        df = df.drop(columns=add_cols)

    return df


def get_test_file(
    fname, base_url="https://cyclobs.ifremer.fr/static/sarwing_datarmor/xsardata"
):
    """
    get test file from base_url(https://cyclobs.ifremer.fr/static/sarwing_datarmor/xsardata/)
    file is unzipped and extracted to `config['data_dir']`

    Parameters
    ----------
    fname: str
        file name to get (without '.zip' extension)

    Returns
    -------
    str
        path to file, relative to `config['data_dir']`

    """
    res_path = config["data_dir"]
    file_url = "%s/%s.zip" % (base_url, fname)
    if not os.path.exists(os.path.join(res_path, fname)):
        warnings.warn("Downloading %s" % file_url)
        local_file = url_get(file_url)
        warnings.warn("Unzipping %s" % os.path.join(res_path, fname))
        with zipfile.ZipFile(local_file, "r") as zip_ref:
            zip_ref.extractall(res_path)
    return os.path.join(res_path, fname)



import xarray as xr
import xsar
import rasterio
import warnings
import os


class XsarXarrayBackend(xr.backends.common.BackendEntrypoint):
    def open_dataset(
        self,
        dataset_id,
        resolution=None,
        resampling=rasterio.enums.Resampling.rms,
        luts=False,
        dtypes=None,
        drop_variables=[],
    ):
        ds = xsar.open_dataset(
            dataset_id,
            resolution=resolution,
            resampling=resampling,
            luts=luts,
            dtypes=dtypes,
        )
        if not list(ds.chunks):
            warnings.warn("Not using `chunks` kw is discouraged when openning SAFE")
        return ds.drop_vars(drop_variables, errors="ignore")

    def guess_can_open(self, filename_or_obj):
        if isinstance(filename_or_obj, xsar.Sentinel1Meta):
            return True
        if isinstance(filename_or_obj, str) and os.path.basename(
            filename_or_obj
        ).endswith(".SAFE"):
            return True
        return False



import rasterio
from rasterio.control import GroundControlPoint
from scipy.interpolate import RectBivariateSpline
from shapely.geometry import Polygon

from .utils import haversine, timing
from .base_meta import BaseMeta
import os
import numpy as np
import pandas as pd
import geopandas as gpd


class RcmMeta(BaseMeta):
    """
    Handle dataset metadata.
    A `xsar.RadarSat2Meta` object can be used with `xsar.open_dataset`,
    but it can be used as itself: it contains usefull attributes and methods.

    Parameters
    ----------
    name: str
        path or gdal identifier

    """

    dt = None

    @timing
    def __init__(self, name):
        super().__init__()

        from safe_rcm import api

        if ":" in name:
            self.dt = api.open_rcm(name.split(":")[1])
        else:
            self.dt = api.open_rcm(name)
        if not name.startswith("RCM_DS:"):
            name = "RCM_DS:%s:" % name
        self.name = name
        """Gdal dataset name"""
        name_parts = self.name.split(":")
        if len(name_parts) > 3:
            # windows might have semicolon in path ('c:\...')
            name_parts[1] = ":".join(name_parts[1:-1])
            del name_parts[2:-1]
        name_parts[1] = os.path.basename(name_parts[1])
        self.short_name = ":".join(name_parts)
        """Like name, but without path"""
        self.path = ":".join(self.name.split(":")[1:-1])
        """Dataset path"""
        self.safe = os.path.basename(self.path)
        """Safe file name"""
        # there is no information on resolution 'F' 'H' or 'M' in the manifest, so we have to extract it from filename
        try:
            self.product = os.path.basename(self.path).split("_")[9]
        except ValueError:
            self.product = "XXX"
        """Product type, like 'GRDH', 'SLC', etc .."""

        # define important xpath to easier access some dataset of the reader's datatree
        self._xpath = {
            "geolocationGrid": "imageReferenceAttributes/geographicInformation/geolocationGrid",
            "lut": "lookupTables/lookupTables",
            "noise_lut": "lookupTables/noiseLevels/referenceNoiseLevel",
            "incidence": "lookupTables/incidenceAngles",
            "attitude": "sourceAttributes/orbitAndAttitude/attitudeInformation",
            "orbit": "sourceAttributes/orbitAndAttitude/orbitInformation",
        }

        # replace `coefficients` dim by `pixel` in the datatree
        var_with_coeff = ["noise_lut", "lut", "incidence"]
        for var in var_with_coeff:
            self.dt[self._xpath[var]] = self.dt[self._xpath[var]].ds.rename(
                {"coefficients": "pixel"}
            )
        # build pixels coords
        self.assign_index(var_with_coeff)

        # flip line and samples when necessary to respect `xsar` convention
        self.samples_flipped = False
        self.lines_flipped = False
        self.flip_sample_da()
        self.flip_line_da()

        self._safe_files = None
        self.multidataset = False
        """True if multi dataset"""
        self.subdatasets = gpd.GeoDataFrame(geometry=[], index=[])
        """Subdatasets as GeodataFrame (empty if single dataset)"""
        self.geoloc = self.dt[self._xpath["geolocationGrid"]].to_dataset()

        self.orbit = self.dt[self._xpath["orbit"]].ds
        self.attitude = self.dt[self._xpath["attitude"]].ds
        self.lut = self.dt[self._xpath["lut"]].ds
        self.noise_lut = self.dt[self._xpath["noise_lut"]].ds
        self.incidence = self.dt[self._xpath["incidence"]].ds
        self.manifest_attrs = self._create_manifest_attrs()
        for name, feature in self.__class__._mask_features_raw.items():
            self.set_mask_feature(name, feature)

    def flip_sample_da(self):
        """
        When a product is flipped, flip back data arrays (from the reader datatree) sample dimensions to respect the xsar
        convention (increasing incidence values).
        Documentation reference : RCM Image Product Format Definition (4.2.1)
        """
        antenna_pointing = self.dt["sourceAttributes/radarParameters"].attrs[
            "antennaPointing"
        ]
        pass_direction = self.dt[
            "sourceAttributes/orbitAndAttitude/orbitInformation"
        ].attrs["passDirection"]
        flipped_cases = [("Left", "Ascending"), ("Right", "Descending")]
        samples_depending_ds = ["geolocationGrid"]
        if (antenna_pointing, pass_direction) in flipped_cases:
            for ds_name in samples_depending_ds:
                self.dt[self._xpath[ds_name]] = (
                    self.dt[self._xpath[ds_name]]
                    .ds.copy()
                    .isel(pixel=slice(None, None, -1))
                    .assign_coords(pixel=self.dt[self._xpath[ds_name]].ds.pixel)
                )
            self.samples_flipped = True
        return

    def flip_line_da(self):
        """
        Flip dataArrays (from the reader datatree) that depend on line dimension when a product is ascending, in order to
        respect the xsar convention (increasing time along line axis, whatever ascending or descending product).
        Documentation reference : RCM Image Product Format Definition (4.2.1)
        """
        pass_direction = self.dt[
            "sourceAttributes/orbitAndAttitude/orbitInformation"
        ].attrs["passDirection"]
        samples_depending_ds = ["geolocationGrid"]
        if pass_direction == "Ascending":
            for ds_name in samples_depending_ds:
                self.dt[self._xpath[ds_name]] = (
                    self.dt[self._xpath[ds_name]]
                    .ds.copy()
                    .isel(line=slice(None, None, -1))
                    .assign_coords(line=self.dt[self._xpath[ds_name]].ds.line)
                )
            self.lines_flipped = True
        return

    def assign_index(self, ds_list):
        """
        Generate a list of increasing indexes as a dataset coord. The change is made directly in the reader's datatree
        to all the datasets given in ds_list.
        The list of indexes is built thanks to a firstPixelValue, a step and a number of values.
        Documentation reference : RCM Image Product Format Definition (7.5.1)

        Parameters
        ----------
        ds_list : List[str]
            List of dataset that don't have the samples list built. The list can contain these values :
            ['lut', 'noise_lut', 'incidence'].

        Returns
        -------
        """

        def _assign_index_lut():
            xpath = self._xpath["lut"]
            lut = self.dt[xpath].ds.lookup_tables
            first_pix = lut.attrs["pixelFirstLutValue"]
            step = lut.attrs["stepSize"]
            values_nb = lut.attrs["numberOfValues"]
            indexes = [first_pix + step * i for i in range(0, values_nb)]
            # we want increasing indexes
            if step < 0:
                indexes = indexes[::-1]
            self.dt[xpath] = self.dt[xpath].ds.assign_coords(pixel=indexes)
            return

        def _assign_index_noise_lut():
            xpath = self._xpath["noise_lut"]
            noise_lut = self.dt[xpath].ds
            first_pix = noise_lut.pixelFirstNoiseValue
            step = noise_lut.stepSize
            values_nb = noise_lut.attrs["numberOfValues"]
            indexes = [first_pix + step * i for i in range(0, values_nb)]
            # we want increasing indexes
            if step < 0:
                indexes = indexes[::-1]
            self.dt[xpath] = self.dt[xpath].ds.assign_coords(pixel=indexes)
            return

        def _assign_index_incidence():
            xpath = self._xpath["incidence"]
            incidence = self.dt[xpath].ds
            first_pix = incidence.attrs["pixelFirstAnglesValue"]
            step = incidence.attrs["stepSize"]
            values_nb = incidence.attrs["numberOfValues"]
            indexes = [first_pix + step * i for i in range(0, values_nb)]
            # we want increasing indexes
            if step < 0:
                indexes = indexes[::-1]
            self.dt[xpath] = self.dt[xpath].ds.assign_coords(pixel=indexes)
            return

        if not all(item in ["lut", "noise_lut", "incidence"] for item in ds_list):
            raise ValueError(
                "Please enter accepted dataset names ('lut', 'noise_lut', 'incidence')"
            )
        if "lut" in ds_list:
            _assign_index_lut()
        if "noise_lut" in ds_list:
            _assign_index_noise_lut()
        if "incidence" in ds_list:
            _assign_index_incidence()
        return

    def _create_manifest_attrs(self):
        dic = dict()
        dic["swath_type"] = os.path.basename(self.path).split("_")[4]
        dic["polarizations"] = self.dt["sourceAttributes/radarParameters"][
            "pole"
        ].values
        dic["product_type"] = self.product
        dic["satellite"] = "RCM"
        dic["start_date"] = self.start_date
        dic["stop_date"] = self.stop_date
        # compute attributes (footprint, coverage, pixel_size)
        footprint_dict = {}
        for ll in ["longitude", "latitude"]:
            footprint_dict[ll] = [
                self.geoloc[ll].isel(line=a, pixel=x).values
                for a, x in [(0, 0), (0, -1), (-1, -1), (-1, 0)]
            ]
        corners = list(zip(footprint_dict["longitude"], footprint_dict["latitude"]))
        p = Polygon(corners)
        self.geoloc.attrs["footprint"] = p
        dic["footprints"] = p
        # compute acquisition size/resolution in meters
        # first vector is on sample
        acq_sample_meters, _ = haversine(*corners[0], *corners[1])
        # second vector is on line
        acq_line_meters, _ = haversine(*corners[1], *corners[2])
        dic["coverage"] = "%dkm * %dkm (line * sample )" % (
            acq_line_meters / 1000,
            acq_sample_meters / 1000,
        )

        def _to_rio_gcp(pt_geoloc):
            # convert a point from self._geoloc grid to rasterio GroundControlPoint
            return GroundControlPoint(
                x=pt_geoloc.longitude.item(),
                y=pt_geoloc.latitude.item(),
                z=pt_geoloc.height.item(),
                col=pt_geoloc.line.item(),
                row=pt_geoloc.pixel.item(),
            )

        gcps = [
            _to_rio_gcp(self.geoloc.sel(line=line, pixel=sample))
            for line in self.geoloc.line
            for sample in self.geoloc.pixel
        ]
        # approx transform, from all gcps (inaccurate)
        dic["approx_transform"] = rasterio.transform.from_gcps(gcps)
        return dic

    @property
    def approx_transform(self):
        """
        Affine transfom from geoloc.

        This is an inaccurate transform, with errors up to 600 meters.
        But it's fast, and may fit some needs, because the error is stable localy.
        See `xsar.BaseMeta.coords2ll` `xsar.BaseMeta.ll2coords` for accurate methods.

        Examples
        --------
            get `longitude` and `latitude` from tuple `(line, sample)`:

            >>> longitude, latitude = self.approx_transform * (line, sample)

            get `line` and `sample` from tuple `(longitude, latitude)`

            >>> line, sample = ~self.approx_transform * (longitude, latitude)

        See Also
        --------
        xsar.BaseDataset.coords2ll
        xsar.BaseDataset.ll2coords
        xsar.BaseMeta.coords2ll
        xsar.BaseMeta.ll2coords

        """
        return self.manifest_attrs["approx_transform"]

    @property
    def footprint(self):
        """footprint, as a shapely polygon or multi polygon"""
        return self.geoloc.attrs["footprint"]

    @property
    def pixel_line_m(self):
        """pixel line spacing, in meters (at sensor level)"""
        if self.multidataset:
            res = None  # not defined for multidataset
        else:
            res = self.dt["imageReferenceAttributes/rasterAttributes"][
                "sampledLineSpacing"
            ].values
        return res

    @property
    def pixel_sample_m(self):
        """pixel sample spacing, in meters (at sensor level)"""
        if self.multidataset:
            res = None  # not defined for multidataset
        else:
            res = self.dt["imageReferenceAttributes/rasterAttributes"][
                "sampledPixelSpacing"
            ].values
        return res

    @property
    def get_azitime(self):
        """
        Get time at low resolution

        Returns
        -------
        array[datetime64[ns]]
            times
        """
        return self.orbit.timeStamp.values

    @property
    def pols(self):
        """polarisations strings, separated by spaces"""
        return " ".join(self.manifest_attrs["polarizations"])

    def __reduce__(self):
        # make self serializable with pickle
        # https://docs.python.org/3/library/pickle.html#object.__reduce__

        return self.__class__, (self.name,), self.dict

    def __repr__(self):
        if self.multidataset:
            meta_type = "multi (%d)" % len(self.subdatasets)
        else:
            meta_type = "single"
        return "<RcmMeta %s object>" % meta_type

    def _get_time_range(self):
        if self.multidataset:
            time_range = [
                self.manifest_attrs["start_date"],
                self.manifest_attrs["stop_date"],
            ]
        else:
            time_range = self.orbit.timeStamp
        return pd.Interval(
            left=pd.Timestamp(time_range.values[0]),
            right=pd.Timestamp(time_range.values[-1]),
            closed="both",
        )

    @property
    def _dict_coords2ll(self):
        """
        dict with keys ['longitude', 'latitude'] with interpolation function (RectBivariateSpline) as values.

        Examples:
        ---------
            get longitude at line=100 and sample=200:
            ```
            >>> self._dict_coords2ll["longitude"].ev(100, 200)
            array(-66.43947434)
            ```
        Notes:
        ------
            if self.cross_antemeridian is True, 'longitude' will be in range [0, 360]
        """
        resdict = {}
        geoloc = self.geoloc
        if self.cross_antemeridian:
            geoloc["longitude"] = geoloc["longitude"] % 360

        idx_sample = np.array(geoloc.pixel)
        idx_line = np.array(geoloc.line)

        for ll in ["longitude", "latitude"]:
            resdict[ll] = RectBivariateSpline(
                idx_line, idx_sample, np.asarray(geoloc[ll]), kx=1, ky=1
            )

        return resdict

    def to_dict(self, keys="minimal"):

        info_keys = {
            "minimal": [
                # 'platform',
                "swath",
                "product",
                "pols",
            ]
        }
        info_keys["all"] = info_keys["minimal"] + [
            "name",
            "start_date",
            "stop_date",
            "footprint",
            "coverage",
            "pixel_line_m",
            "pixel_sample_m",
            "approx_transform",
            # 'orbit_pass',
            # 'platform_heading'
        ]

        if isinstance(keys, str):
            keys = info_keys[keys]

        res_dict = {}
        for k in keys:
            if hasattr(self, k):
                res_dict[k] = getattr(self, k)
            elif k in self.manifest_attrs.keys():
                res_dict[k] = self.manifest_attrs[k]
            else:
                raise KeyError('Unable to find key/attr "%s" in RcmMeta' % k)
        return res_dict



# -*- coding: utf-8 -*-
import glob
import logging
import os
import warnings

import rasterio
import yaml
from affine import Affine

from .rcm_meta import RcmMeta
import numpy as np
from .utils import timing, map_blocks_coords, BlockingActorProxy, to_lon180, get_glob
from scipy.interpolate import RectBivariateSpline, interp1d
import dask
import xarray as xr
import rioxarray
from .base_dataset import BaseDataset

logger = logging.getLogger("xsar.radarsat2_dataset")
logger.addHandler(logging.NullHandler())

# we know tiff as no geotransform : ignore warning
warnings.filterwarnings("ignore", category=rasterio.errors.NotGeoreferencedWarning)

# allow nan without warnings
# some dask warnings are still non filtered: https://github.com/dask/dask/issues/3245
np.errstate(invalid="ignore")


class RcmDataset(BaseDataset):
    """
    Handle a SAFE subdataset.
    A dataset might contain several tiff files (multiples polarizations), but all tiff files must share the same footprint.

    The main attribute useful to the end-user is `self.dataset` (`xarray.Dataset` , with all variables parsed from xml and tiff files.)

    Parameters
    ----------
    dataset_id: str or RadarSat2Meta object

        if str, it can be a path, or a gdal dataset identifier)

    resolution: dict, number or string, optional
        resampling dict like `{'line': 20, 'sample': 20}` where 20 is in pixels.

        if a number, dict will be constructed from `{'line': number, 'sample': number}`

        if str, it must end with 'm' (meters), like '100m'. dict will be computed from sensor pixel size.

    resampling: rasterio.enums.Resampling or str, optional

        Only used if `resolution` is not None.

        ` rasterio.enums.Resampling.rms` by default. `rasterio.enums.Resampling.nearest` (decimation) is fastest.

    chunks: dict, optional

        dict with keys ['pol','line','sample'] (dask chunks).

    dtypes: None or dict, optional

        Specify the data type for each variable.

    lazyloading: bool, optional
        activate or not the lazy loading of the high resolution fields

    """

    def __init__(
        self,
        dataset_id,
        resolution=None,
        resampling=rasterio.enums.Resampling.rms,
        chunks={"line": 5000, "sample": 5000},
        dtypes=None,
        lazyloading=True,
        skip_variables=None,
    ):
        if skip_variables is None:
            # TODO : Remove `velocity` from the skip_variable when the problem is resolved
            #  (must understand link between time and lines)
            skip_variables = ["velocity"]
        # default dtypes
        if dtypes is not None:
            self._dtypes.update(dtypes)

        # default meta for map_blocks output.
        # as asarray is imported from numpy, it's a numpy array.
        # but if later we decide to import asarray from cupy, il will be a cupy.array (gpu)
        self.sar_meta = None
        self.resolution = resolution

        if not isinstance(dataset_id, RcmMeta):
            self.sar_meta = BlockingActorProxy(RcmMeta, dataset_id)
            # check serializable
            # import pickle
            # s1meta = pickle.loads(pickle.dumps(self.s1meta))
            # assert isinstance(rs2meta.coords2ll(100, 100),tuple)
        else:
            # we want self.rs2meta to be a dask actor on a worker
            self.sar_meta = BlockingActorProxy(RcmMeta.from_dict, dataset_id.dict)
        del dataset_id

        if self.sar_meta.multidataset:
            raise IndexError(
                """Can't open an multi-dataset. Use `xsar.RadarSat2Meta('%s').subdatasets` to show availables ones"""
                % self.sar_meta.path
            )

        # build datatree
        DN_tmp = self.load_digital_number(
            resolution=resolution, resampling=resampling, chunks=chunks
        )
        DN_tmp = self.flip_sample_da(DN_tmp)
        DN_tmp = self.flip_line_da(DN_tmp)

        # geoloc
        geoloc = self.sar_meta.geoloc
        geoloc.attrs["history"] = "annotations"

        # orbitInformation
        orbit = self.sar_meta.orbit
        orbit.attrs["history"] = "annotations"

        self.datatree = xr.DataTree.from_dict(
            {"measurement": DN_tmp, "geolocation_annotation": geoloc}
        )

        self._dataset = self.datatree["measurement"].to_dataset()

        # merge the datatree with the reader
        for group in self.sar_meta.dt:
            self.datatree[group] = self.sar_meta.dt[group]

        # dict mapping for calibration type in the reader
        self._map_calibration_type = {
            "sigma0": "Sigma Nought",
            "gamma0": "Gamma",
            "beta0": "Beta Nought",
        }

        self._map_var_lut = {
            "sigma0": "sigma0",
            "gamma0": "gamma0",
            "beta0": "beta0",
        }

        geoloc_vars = ["latitude", "longitude", "altitude", "incidence", "elevation"]
        for vv in skip_variables:
            if vv in geoloc_vars:
                geoloc_vars.remove(vv)

        for att in ["name", "short_name", "product", "safe", "swath", "multidataset"]:
            if att not in self.datatree.attrs:
                # tmp = xr.DataArray(self.s1meta.__getattr__(att),attrs={'source':'filename decoding'})
                self.datatree.attrs[att] = self.sar_meta.__getattr__(att)
                self._dataset.attrs[att] = self.sar_meta.__getattr__(att)

        value_res_line = self.sar_meta.pixel_line_m
        value_res_sample = self.sar_meta.pixel_sample_m
        # self._load_incidence_from_lut()
        refe_spacing = "slant"
        if resolution is not None:
            # if the data sampling changed it means that the quantities are projected on ground
            refe_spacing = "ground"
            if isinstance(resolution, str):
                value_res_sample = float(resolution.replace("m", ""))
                value_res_line = value_res_sample
            elif isinstance(resolution, dict):
                value_res_sample = self.sar_meta.pixel_sample_m * resolution["sample"]
                value_res_line = self.sar_meta.pixel_line_m * resolution["line"]
            else:
                logger.warning(
                    "resolution type not handle (%s) should be str or dict -> sampleSpacing"
                    " and lineSpacing are not correct",
                    type(resolution),
                )
        self._dataset["sampleSpacing"] = xr.DataArray(
            value_res_sample,
            attrs={"referential": refe_spacing}
            | self.sar_meta.dt["imageReferenceAttributes/rasterAttributes"][
                "sampledPixelSpacing"
            ].attrs,
        )
        self._dataset["lineSpacing"] = xr.DataArray(
            value_res_line,
            attrs=self.sar_meta.dt["imageReferenceAttributes/rasterAttributes"][
                "sampledLineSpacing"
            ].attrs,
        )

        # dataset no-pol template for function evaluation on coordinates (*no* values used)
        # what's matter here is the shape of the image, not the values.
        with warnings.catch_warnings():
            # warnings.simplefilter("ignore", np.ComplexWarning)
            self._da_tmpl = xr.DataArray(
                dask.array.empty_like(
                    self._dataset.digital_number.isel(pol=0).drop("pol"),
                    dtype=np.int8,
                    name="empty_var_tmpl-%s" % dask.base.tokenize(self.sar_meta.name),
                ),
                dims=("line", "sample"),
                coords={
                    "line": self._dataset.digital_number.line,
                    "sample": self._dataset.digital_number.sample,
                },
            )

            # Add vars to define if lines or samples have been flipped to respect xsar convention
            self._dataset = xr.merge(
                [
                    xr.DataArray(
                        data=self.sar_meta.samples_flipped,
                        attrs={
                            "meaning": "xsar convention : increasing incidence values along samples axis"
                        },
                    ).to_dataset(name="samples_flipped"),
                    self._dataset,
                ]
            )
            self._dataset = xr.merge(
                [
                    xr.DataArray(
                        data=self.sar_meta.lines_flipped,
                        attrs={
                            "meaning": "xsar convention : increasing time along line axis "
                            "(whatever ascending or descending pass direction)"
                        },
                    ).to_dataset(name="lines_flipped"),
                    self._dataset,
                ]
            )

        self._luts = self.lazy_load_luts()
        self._noise_luts = self.lazy_load_noise_luts()
        self._noise_luts = self._noise_luts.drop(["pixelFirstNoiseValue", "stepSize"])
        self.apply_calibration_and_denoising()
        self._dataset = xr.merge(
            [
                self.load_from_geoloc(geoloc_vars, lazy_loading=lazyloading),
                self._dataset,
            ]
        )
        # compute offboresight in self._dataset
        self._get_offboresight_from_elevation()

        rasters = self._load_rasters_vars()
        if rasters is not None:
            self._dataset = xr.merge([self._dataset, rasters])
        if "ground_heading" not in skip_variables:
            self._dataset = xr.merge([self.load_ground_heading(), self._dataset])
        if "velocity" not in skip_variables:
            self._dataset = xr.merge([self.get_sensor_velocity(), self._dataset])
        self._rasterized_masks = self.load_rasterized_masks()
        self._dataset = xr.merge([self._rasterized_masks, self._dataset])
        self.datatree["measurement"] = self.datatree["measurement"].assign(
            self._dataset
        )
        # merge the datatree with the reader

        self.reconfigure_reader_datatree()
        self._dataset.attrs.update(self.sar_meta.to_dict("all"))
        self.datatree.attrs.update(self.sar_meta.to_dict("all"))

    def lazy_load_luts(self):
        """
        Lazy load luts from the reader as delayed

        Returns
        -------
        xarray.Dataset
            Contains delayed dataArrays of luts
        """
        merge_list = []
        for key, value in self._map_calibration_type.items():
            list_da = []
            for pola in self.sar_meta.lut.lookup_tables.pole:
                lut = self.sar_meta.lut.lookup_tables.sel(
                    sarCalibrationType=value, pole=pola
                ).rename(
                    {
                        "pixel": "sample",
                    }
                )
                values_nb = lut.attrs["numberOfValues"]
                lut_f_delayed = dask.delayed()(lut)
                ar = dask.array.from_delayed(
                    lut_f_delayed.data, (values_nb,), lut.dtype
                )
                da = xr.DataArray(
                    data=ar,
                    dims=["sample"],
                    coords={"sample": lut.sample},
                    attrs=lut.attrs,
                )
                da = (
                    self._interpolate_var(da)
                    .assign_coords(pole=pola)
                    .rename({"pole": "pol"})
                )
                list_da.append(da)
            full_da = xr.concat(list_da, dim="pol")
            full_da.attrs = lut.attrs
            ds_lut_f_delayed = full_da.to_dataset(name=key)
            merge_list.append(ds_lut_f_delayed)
        return xr.merge(merge_list)

    def lazy_load_noise_luts(self):
        """
        Lazy load noise luts from the reader as delayed

        Returns
        -------
        xarray.Dataset
            Contains delayed dataArrays of luts
        """
        merge_list = []
        for key, value in self._map_calibration_type.items():
            list_da = []
            values_nb = self.sar_meta.noise_lut.attrs["numberOfValues"]
            for pola in self.sar_meta.noise_lut.noiseLevelValues.pole:
                lut = self.sar_meta.noise_lut.noiseLevelValues.sel(
                    sarCalibrationType=value, pole=pola
                ).rename(
                    {
                        "pixel": "sample",
                    }
                )
                lut_f_delayed = dask.delayed()(lut)
                ar = dask.array.from_delayed(
                    lut_f_delayed.data, (values_nb,), lut.dtype
                )
                da = xr.DataArray(
                    data=ar,
                    dims=["sample"],
                    coords={"sample": lut.sample},
                    attrs=lut.attrs,
                )
                da = (
                    self._interpolate_var(da, type="noise")
                    .assign_coords(pole=pola)
                    .rename({"pole": "pol"})
                )
                list_da.append(da)
            full_da = xr.concat(list_da, dim="pol")
            ds_lut_f_delayed = full_da.to_dataset(name=key)
            merge_list.append(ds_lut_f_delayed)
        return xr.merge(merge_list)

    @timing
    def _interpolate_var(self, var, type="lut"):
        """
        Interpolate look up table (from the reader) or another variable and resample it.
        Initial values are at low resolution, and the high resolution range is made from the pixel first noise
        level value and the step. Then, an interpolation with RectBivariateSpline permit having a full resolution
        and extrapolate the first pixels; getting by the end resampled look up tables.

        Parameters
        ----------
        var: xarray.DataArray
            variable we want to interpolate (lut extracted from the reader :noise or calibration ; incidence; elevation)

        type: str
            type of variable we want to interpolate. Can be "lut", "noise", "incidence"

        Returns
        -------
        xarray.DataArray
            Variable interpolated and resampled
        """
        accepted_types = ["lut", "noise", "incidence"]
        if type not in accepted_types:
            raise ValueError(
                "Please enter a type accepted ('lut', 'noise', 'incidence')"
            )
        lines = self.sar_meta.geoloc.line
        samples = var.sample
        var_type = None
        if type == "noise":
            # Give the good saving type and convert the noise values to linear
            var_type = self._dtypes["noise_lut"]
            var = 10 ** (var / 10)
        elif type == "lut":
            var_type = self._dtypes["sigma0_lut"]
        elif type == "incidence":
            var_type = self._dtypes[type]
        var_2d = np.tile(var, (lines.shape[0], 1))
        interp_func = dask.delayed(RectBivariateSpline)(
            x=lines, y=samples, z=var_2d, kx=1, ky=1
        )
        da_var = map_blocks_coords(self._da_tmpl.astype(var_type), interp_func)
        return da_var

    @timing
    def _apply_calibration_lut(self, var_name):
        """
        Apply calibration lut to `digital_number` to compute `var_name`.

        Parameters
        ----------
        var_name: str
            Variable name to compute by applying lut. Must exist in `self._map_var_lut` to be able to get the corresponding lut.

        Returns
        -------
        xarray.Dataset
            with one variable named by `var_name`
        """
        lut = self._get_lut(var_name)
        offset = lut.attrs["offset"]
        res = ((self._dataset.digital_number**2.0) + offset) / lut
        res = res.where(res > 0)
        res.attrs.update(lut.attrs)
        return res.to_dataset(name=var_name + "_raw")

    @timing
    def _get_noise(self, var_name):
        """
        Get noise equivalent for  `var_name`.

        Parameters
        ----------
        var_name: str Variable name to compute. Must exist in `self._map_var_lut` and
        `self._map_var_lut_noise` to be able to get the corresponding lut.

        Returns
        -------
        xarray.Dataset
            with one variable named by `'ne%sz' % var_name[0]` (ie 'nesz' for 'sigma0', 'nebz' for 'beta0', etc...)
        """
        name = "ne%sz" % var_name[0]
        try:
            lut_name = self._map_var_lut[var_name]
        except KeyError:
            raise ValueError("can't find lut name for var '%s'" % var_name)
        try:
            lut = self._noise_luts[lut_name]
        except KeyError:
            raise ValueError(
                "can't find noise lut from name '%s' for variable '%s' "
                % (lut_name, var_name)
            )
        return lut.to_dataset(name=name)

    def apply_calibration_and_denoising(self):
        """
        apply calibration and denoising functions to get high resolution sigma0 , beta0 and gamma0 + variables *_raw

        Returns:
        --------

        """
        for var_name, lut_name in self._map_var_lut.items():
            if lut_name in self._luts:
                # merge var_name into dataset (not denoised)
                self._dataset = self._dataset.merge(
                    self._apply_calibration_lut(var_name)
                )
                # merge noise equivalent for var_name (named 'ne%sz' % var_name[0)
                self._dataset = self._dataset.merge(self._get_noise(var_name))
            else:
                logger.debug(
                    "Skipping variable '%s' ('%s' lut is missing)"
                    % (var_name, lut_name)
                )
        self._dataset = self._add_denoised(self._dataset)
        self.datatree["measurement"] = self.datatree["measurement"].assign(
            self._dataset
        )
        # self._dataset = self.datatree[
        #     'measurement'].to_dataset()  # test oct 22 to see if then I can modify variables of the dt
        return

    def _add_denoised(self, ds, clip=False, vars=None):
        """add denoised vars to dataset

        Parameters
        ----------
        ds : xarray.DataSet
            dataset with non denoised vars, named `%s_raw`.
        clip : bool, optional
            If True, negative signal will be clipped to 0. (default to False )
        vars : list, optional
            variables names to add, by default `['sigma0' , 'beta0' , 'gamma0']`

        Returns
        -------
        xarray.DataSet
            dataset with denoised vars
        """
        already_denoised = self.datatree["imageGenerationParameters"][
            "sarProcessingInformation"
        ].attrs["noiseSubtractionPerformed"]

        if vars is None:
            vars = ["sigma0", "beta0", "gamma0"]
        for varname in vars:

            varname_raw = varname + "_raw"
            noise = "ne%sz" % varname[0]
            if varname_raw not in ds:
                continue
            else:

                if not already_denoised:
                    denoised = ds[varname_raw] - ds[noise]
                    if clip:
                        denoised = denoised.clip(min=0)
                        denoised.attrs["comment"] = "clipped, no values <0"
                    else:
                        denoised.attrs["comment"] = "not clipped, some values can be <0"
                    ds[varname] = denoised

                else:
                    logging.debug(
                        "product was already denoised (noiseSubtractionPerformed = True), noise added back"
                    )
                    denoised = ds[varname_raw]
                    denoised.attrs["denoising information"] = (
                        "already denoised by Canadian Agency"
                    )
                    if clip:
                        denoised = denoised.clip(min=0)
                        denoised.attrs["comment"] = "clipped, no values <0"
                    else:
                        denoised.attrs["comment"] = "not clipped, some values can be <0"
                    ds[varname] = denoised

                    ds[varname_raw] = denoised + ds[noise]
                    ds[varname_raw].attrs[
                        "denoising information"
                    ] = "product was already denoised, noise added back"

        return ds

    def _load_incidence_from_lut(self):
        """
        Load incidence from the reader as delayed

        Returns
        -------
        xarray.Dataset
            Contains delayed dataArrays of incidence
        """
        incidence = self.sar_meta.incidence.rename({"pixel": "sample"})
        angles = incidence.angles
        values_nb = incidence.attrs["numberOfValues"]
        lut_f_delayed = dask.delayed()(angles)
        ar = dask.array.from_delayed(
            lut_f_delayed.data, (values_nb,), self._dtypes["incidence"]
        )
        da = xr.DataArray(
            data=ar,
            dims=["sample"],
            coords={"sample": angles.sample},
            attrs=angles.attrs,
        )
        da = self._interpolate_var(da, type="incidence")
        # ds_lut_f_delayed = da.to_dataset(name='incidence')
        # ds_lut_f_delayed.attrs = incidence.attrs
        return da

    @timing
    def _load_elevation_from_lut(self):
        """
        Load elevation from lut.
        this formula needs the orbit altitude. But 2 variables look like this one : `satelliteHeight` and `Altitude`.
        We considered the satelliteHeight.

        Returns
        -------

        """
        satellite_height = (
            self.sar_meta.dt["imageGenerationParameters/sarProcessingInformation"]
            .to_dataset()
            .satelliteHeight
        )
        earth_radius = 6.371e6
        incidence = self._load_incidence_from_lut()
        angle_rad = np.sin(np.radians(incidence))
        inside = angle_rad * earth_radius / (earth_radius + satellite_height)
        return np.degrees(np.arcsin(inside))

    @timing
    def _get_offboresight_from_elevation(self):
        """
        Compute offboresight angle.

        Returns
        -------

        """
        self._dataset["offboresight"] = (
            self._dataset.elevation
            - (
                30.1833947 * self._dataset.latitude**0
                + 0.0082998714 * self._dataset.latitude**1
                - 0.00031181534 * self._dataset.latitude**2
                - 0.0943533e-07 * self._dataset.latitude**3
                + 3.0191435e-08 * self._dataset.latitude**4
                + 4.968415428e-12 * self._dataset.latitude**5
                - 9.315371305e-13 * self._dataset.latitude**6
            )
            + 29.45
        )
        self._dataset["offboresight"].attrs[
            "comment"
        ] = "built from elevation angle and latitude"

    @timing
    def load_from_geoloc(self, varnames, lazy_loading=True):
        """
        Interpolate (with RectBiVariateSpline) variables from `self.sar_meta.geoloc` to `self._dataset`

        Parameters
        ----------
        varnames: list of str
            subset of variables names in `self.sar_meta.geoloc`

        Returns
        -------
        xarray.Dataset
            With interpolated variables

        """
        mapping_dataset_geoloc = {
            "latitude": "latitude",
            "longitude": "longitude",
            "incidence": "incidenceAngle",
            "elevation": "elevationAngle",
            "altitude": "height",
        }
        da_list = []

        for varname in varnames:
            varname_in_geoloc = mapping_dataset_geoloc[varname]

            if varname == "incidence":
                da = self._load_incidence_from_lut()
                da.name = varname
                da_list.append(da)
            elif varname == "elevation":
                da = self._load_elevation_from_lut()
                da.name = varname
                da_list.append(da)

            else:
                if varname == "longitude":
                    z_values = self.sar_meta.geoloc[varname]
                    if self.sar_meta.cross_antemeridian:
                        logger.debug("translate longitudes between 0 and 360")
                        z_values = z_values % 360
                else:
                    z_values = self.sar_meta.geoloc[varname_in_geoloc]
                interp_func = RectBivariateSpline(
                    self.sar_meta.geoloc.line,
                    self.sar_meta.geoloc.pixel,
                    z_values,
                    kx=1,
                    ky=1,
                )
                typee = self.sar_meta.geoloc[varname_in_geoloc].dtype
                if lazy_loading:
                    da_var = map_blocks_coords(self._da_tmpl.astype(typee), interp_func)
                else:
                    da_val = interp_func(
                        self._dataset.digital_number.line,
                        self._dataset.digital_number.sample,
                    )
                    da_var = xr.DataArray(
                        data=da_val,
                        dims=["line", "sample"],
                        coords={
                            "line": self._dataset.digital_number.line,
                            "sample": self._dataset.digital_number.sample,
                        },
                    )
                if varname == "longitude":
                    if self.sar_meta.cross_antemeridian:
                        da_var.data = da_var.data.map_blocks(to_lon180)

                da_var.name = varname

                # copy history
                try:
                    da_var.attrs["history"] = self.sar_meta.geoloc[
                        varname_in_geoloc
                    ].attrs["xpath"]
                except KeyError:
                    pass

                da_list.append(da_var)
        ds = xr.merge(da_list)
        return ds

    @property
    def interpolate_times(self):
        """
        Apply interpolation with RectBivariateSpline to the azimuth time extracted from `self.sar_meta.geoloc`

        Returns
        -------
        xarray.Dataset
            Contains the time as delayed at the good resolution and expressed as type datetime64[ns]
        """
        times = self.sar_meta.get_azitime
        lines = self.sar_meta.geoloc.line
        samples = self.sar_meta.geoloc.pixel
        time_values_2d = np.tile(times, (samples.shape[0], 1)).transpose()
        interp_func = RectBivariateSpline(
            x=lines, y=samples, z=time_values_2d.astype(float), kx=1, ky=1
        )
        da_var = map_blocks_coords(self._da_tmpl.astype("datetime64[ns]"), interp_func)
        return da_var.isel(sample=0).to_dataset(name="time")

    def get_sensor_velocity(self):
        """
        Interpolated sensor velocity
        Returns
        -------
        xarray.Dataset()
            containing a single variable velocity
        """
        azimuth_times = self.sar_meta.get_azitime
        orbstatevect = self.sar_meta.orbit
        velos = np.array(
            [
                orbstatevect["xVelocity"] ** 2.0,
                orbstatevect["yVelocity"] ** 2.0,
                orbstatevect["zVelocity"] ** 2.0,
            ]
        )
        vels = np.sqrt(np.sum(velos, axis=0))
        interp_f = interp1d(azimuth_times.astype(float), vels)
        _vels = interp_f(self.interpolate_times["time"].astype(float))
        res = xr.DataArray(_vels, dims=["line"], coords={"line": self.dataset.line})
        return xr.Dataset({"velocity": res})

    @timing
    def load_digital_number(
        self, resolution=None, chunks=None, resampling=rasterio.enums.Resampling.rms
    ):
        """
        load digital_number from tiff files, as an `xarray.Dataset`.

        Parameters
        ----------
        resolution: None, number, str or dict
            see `xsar.open_dataset`
        resampling: rasterio.enums.Resampling
            see `xsar.open_dataset`

        Returns
        -------
        xarray.Dataset
            dataset (possibly dual-pol), with basic coords/dims naming convention
        """

        def _list_tiff_files():
            """
            Return a list that contains all tiff files paths

            Returns
            -------
            List[str]
                List of Tiff file paths located in a folder
            """

            return glob.glob(os.path.join(self.sar_meta.path, "imagery", "*.tif"))

        def _sort_list_files_and_get_pols(list_tiff):
            """
            From a list of tiff files, sort it to get the co polarization tiff file as the first element, and extract pols

            Parameters
            ----------
            list_tiff: List[str]
                List of tiff files

            Returns
            -------
            (List[str], List[str])
                Tuple that contains the tiff files list sorted and the polarizations
            """
            pols = []
            if len(list_tiff) > 1:
                first_base = os.path.basename(list_tiff[0]).split(".")[0]
                first_pol = first_base[-2:]
                if first_pol[0] != first_pol[1]:
                    list_tiff.reverse()
            for file in list_tiff:
                base = os.path.basename(file).split(".")[0]
                pol = base[-2:]
                pols.append(pol)
            return list_tiff, pols

        map_dims = {"pol": "band", "line": "y", "sample": "x"}
        tiff_files = _list_tiff_files()
        tiff_files, pols = _sort_list_files_and_get_pols(tiff_files)
        if resolution is not None:
            comment = 'resampled at "%s" with %s.%s.%s' % (
                resolution,
                resampling.__module__,
                resampling.__class__.__name__,
                resampling.name,
            )
        else:
            comment = "read at full resolution"

        # arbitrary rio object, to get shape, etc ... (will not be used to read data)
        rio = rasterio.open(tiff_files[0])

        chunks["pol"] = 1
        # sort chunks keys like map_dims
        chunks = dict(
            sorted(
                chunks.items(), key=lambda pair: list(map_dims.keys()).index(pair[0])
            )
        )
        chunks_rio = {map_dims[d]: chunks[d] for d in map_dims.keys()}
        self.resolution = None
        if resolution is None:
            # using tiff driver: need to read individual tiff and concat them
            # riofiles['rio'] is ordered like self.s1meta.manifest_attrs['polarizations']

            dn = xr.concat(
                [
                    rioxarray.open_rasterio(
                        f, chunks=chunks_rio, parse_coordinates=False
                    )
                    for f in tiff_files
                ],
                "band",
            ).assign_coords(band=np.arange(len(pols)) + 1)

            # set dimensions names
            dn = dn.rename(dict(zip(map_dims.values(), map_dims.keys())))

            # create coordinates from dimension index (because of parse_coordinates=False)
            dn = dn.assign_coords({"line": dn.line, "sample": dn.sample})
            dn = dn.drop_vars("spatial_ref", errors="ignore")
        else:
            if not isinstance(resolution, dict):
                if isinstance(resolution, str) and resolution.endswith("m"):
                    resolution = float(resolution[:-1])
                    self.resolution = resolution
                resolution = dict(
                    line=resolution / self.sar_meta.pixel_line_m,
                    sample=resolution / self.sar_meta.pixel_sample_m,
                )
                # resolution = dict(line=resolution / self.dataset['sampleSpacing'].values,
                #                   sample=resolution / self.dataset['lineSpacing'].values)

            # resample the DN at gdal level, before feeding it to the dataset
            out_shape = (
                int(rio.height / resolution["line"]),
                int(rio.width / resolution["sample"]),
            )
            out_shape_pol = (1,) + out_shape
            # read resampled array in one chunk, and rechunk
            # this doesn't optimize memory, but total size remain quite small

            if isinstance(resolution["line"], int):
                # legacy behaviour: winsize is the maximum full image size that can be divided  by resolution (int)
                winsize = (
                    0,
                    0,
                    rio.width // resolution["sample"] * resolution["sample"],
                    rio.height // resolution["line"] * resolution["line"],
                )
                window = rasterio.windows.Window(*winsize)
            else:
                window = None

            dn = xr.concat(
                [
                    xr.DataArray(
                        dask.array.from_array(
                            rasterio.open(f).read(
                                out_shape=out_shape_pol,
                                resampling=resampling,
                                window=window,
                            ),
                            chunks=chunks_rio,
                        ),
                        dims=tuple(map_dims.keys()),
                        coords={"pol": [pol]},
                    )
                    for f, pol in zip(tiff_files, pols)
                ],
                "pol",
            ).chunk(chunks)

            # create coordinates at box center
            translate = Affine.translation(
                (resolution["sample"] - 1) / 2, (resolution["line"] - 1) / 2
            )
            scale = Affine.scale(
                rio.width // resolution["sample"] * resolution["sample"] / out_shape[1],
                rio.height // resolution["line"] * resolution["line"] / out_shape[0],
            )
            sample, _ = translate * scale * (dn.sample, 0)
            _, line = translate * scale * (0, dn.line)
            dn = dn.assign_coords({"line": line, "sample": sample})

        # for GTiff driver, pols are already ordered. just rename them
        dn = dn.assign_coords(pol=pols)

        descr = "not denoised"
        var_name = "digital_number"

        dn.attrs = {
            "comment": "%s digital number, %s" % (descr, comment),
            "history": yaml.safe_dump(
                {
                    var_name: get_glob(
                        [p.replace(self.sar_meta.path + "/", "") for p in tiff_files]
                    )
                }
            ),
        }
        ds = dn.to_dataset(name=var_name)
        astype = self._dtypes.get(var_name)
        if astype is not None:
            ds = ds.astype(self._dtypes[var_name])

        return ds

    def __repr__(self):
        if self.sliced:
            intro = "sliced"
        else:
            intro = "full coverage"
        return "<RcmDataset %s object>" % intro

    @timing
    def flip_sample_da(self, ds):
        """
        When a product is flipped, flip back data arrays (from a dataset) sample dimensions to respect the xsar
        convention (increasing incidence values).
        Documentation reference : RCM Image Product Format Definition (4.2.1)

        Parameters
        ----------
        ds : xarray.Dataset
            Contains dataArrays which depends on `sample` dimension

        Returns
        -------
        xarray.Dataset
            Flipped back, respecting the xsar convention
        """
        antenna_pointing = self.sar_meta.dt["sourceAttributes/radarParameters"].attrs[
            "antennaPointing"
        ]
        pass_direction = self.sar_meta.dt[
            "sourceAttributes/orbitAndAttitude/orbitInformation"
        ].attrs["passDirection"]
        flipped_cases = [("Left", "Ascending"), ("Right", "Descending")]
        if (antenna_pointing, pass_direction) in flipped_cases:
            new_ds = (
                ds.copy()
                .isel(sample=slice(None, None, -1))
                .assign_coords(sample=ds.sample)
            )
        else:
            new_ds = ds
        return new_ds

    @timing
    def flip_line_da(self, ds):
        """
        Flip dataArrays (from a dataset) that depend on line dimension when a product is ascending, in order to
        respect the xsar convention (increasing time along line axis, whatever ascending or descending product).
        Documentation reference : RCM Image Product Format Definition (4.2.1)

        Parameters
        ----------
        ds : xarray.Dataset
            Contains dataArrays which depends on `line` dimension

        Returns
        -------
        xarray.Dataset
            Flipped back, respecting the xsar convention
        """
        pass_direction = self.sar_meta.dt[
            "sourceAttributes/orbitAndAttitude/orbitInformation"
        ].attrs["passDirection"]
        if pass_direction == "Ascending":
            new_ds = (
                ds.copy().isel(line=slice(None, None, -1)).assign_coords(line=ds.line)
            )
        else:
            new_ds = ds.copy()
        return new_ds

    def reconfigure_reader_datatree(self):
        """
        Merge attributes of the reader's datatree in the attributes of self.datatree
        """
        self.datatree.attrs |= self.sar_meta.dt.attrs
        return

    @property
    def dataset(self):
        """
        `xarray.Dataset` representation of this `xsar.RcmDataset` object.
        This property can be set with a new dataset, if the dataset was computed from the original dataset.
        """
        # return self._dataset
        res = self.datatree["measurement"].to_dataset()
        res.attrs = self.datatree.attrs
        return res

    @dataset.setter
    def dataset(self, ds):
        if self.sar_meta.name == ds.attrs["name"]:
            # check if new ds has changed coordinates
            if not self.sliced:
                self.sliced = any(
                    [
                        list(ds[d].values) != list(self._dataset[d].values)
                        for d in ["line", "sample"]
                    ]
                )
            self._dataset = ds
            # self._dataset = self.datatree['measurement'].ds
            self.recompute_attrs()
        else:
            raise ValueError("dataset must be same kind as original one.")

    @dataset.deleter
    def dataset(self):
        logger.debug("deleter dataset")



#!/usr/bin/env python

import logging
import argparse
import shutil
import os
import rasterio
import warnings
import xarray as xr
import numpy as np
import subprocess
import glob
from xsar.utils import compress_safe

warnings.filterwarnings("ignore", category=rasterio.errors.NotGeoreferencedWarning)
logging.basicConfig()
logger = logging.getLogger(os.path.basename(__file__))

logging.getLogger('rasterio').setLevel(logging.CRITICAL)


def get_dir_size(path):
    """

    Parameters
    ----------
    path: str

    Returns
    -------
    float
        dir size in Mb

    """

    total_size = 0
    for dirpath, dirnames, filenames in os.walk(path):
        for f in filenames:
            fp = os.path.join(dirpath, f)
            # skip if it is symbolic link
            if not os.path.islink(fp):
                total_size += os.path.getsize(fp)

    return total_size / (1024 * 1024)


def generate_product_id(smooth, constant=None, initial_product_id='0000'):
    """

    Parameters
    ----------
    smooth: int
    constant: None or not None
    initial_product_id: str

    Returns
    -------
    str
        new 4 chars product id.
         - 1st char: 'Z' if smooth >= 1, or 'C' if constant is not Nonee
         - 2-4 chars : smooth size or 00 if constant, in decimal int


    """
    if constant is not None:
        res_id = 'C%0.3d' % constant
    else:
        res_id = 'Z%0.3d' % smooth

    return res_id


if __name__ == '__main__':
    # default smooth size
    smooth_size = {
        'IW': 10,
        'EW': 5,
        'WV': 10
    }

    parser = argparse.ArgumentParser(description='compress safe')
    parser.add_argument('safe_path')
    parser.add_argument('-d', action='store', help='out dir', default='./', required=False)
    parser.add_argument('-z', action='store',
                        help='compress format (see https://rasterio.readthedocs.io/en/latest/api/rasterio.enums.html#rasterio.enums.Compression)',
                        default='zstd', required=False)
    parser.add_argument('-c', action='store', help='fix constant value in band', required=False)
    parser.add_argument('-s', action='store', help='average dn by s*s box (in pixel). Use 0 to keep full resolution',
                        default='auto')
    args = parser.parse_args()

    safe_path = os.path.normpath(os.path.expanduser(args.safe_path))
    safe_name = os.path.basename(safe_path)
    safe_type = safe_name.split('_')[1]
    safe_product = os.path.splitext(safe_name)[0].split('_')[0]
    if args.s == 'auto':
        if safe_type in smooth_size:
            smooth = smooth_size[safe_type]
        else:
            raise NotImplementedError('no default smooth size for %s. Use explicit -s' % safe_type)
    else:
        smooth = int(args.s)

    # in "*_XXXX.SAFE", XXXX is the product id
    product_id_init = os.path.splitext(safe_name)[0].split('_')[-1]
    product_id_out = generate_product_id(smooth)

    safe_path_out = safe_name.replace(product_id_init, product_id_out)
    safe_path_out = os.path.join(args.d, safe_path_out)

    print("Compressing %s..." % os.path.basename(safe_path))
    safe_out = compress_safe(safe_path, safe_path_out, product=safe_product, smooth=smooth,
                             rasterio_kwargs={'compress': args.z})
    in_size = get_dir_size(safe_path)
    out_size = get_dir_size(safe_out)
    ratio = (out_size / in_size) * 100
    zip_file = '%s.zip' % os.path.basename(safe_out)
    zip_msg = subprocess.check_output('cd %s ; zip -r %s %s' % (args.d, zip_file, os.path.basename(safe_out)),
                                      shell=True)
    zip_path = os.path.join(args.d, zip_file)
    zip_size = os.path.getsize(zip_path) / (1024 ** 2)  # in Mb
    zip_ratio = (zip_size / out_size) * 100
    all_ratio = (zip_size / in_size) * 100
    print("%.0fM -> %.0fM ( %.1f%% ) (%s) -> %.0fM ( %.1f%% ) (zip)" % (
        in_size, out_size, ratio, args.z, zip_size, all_ratio))
    # clean temporary path
    subprocess.check_output('cd %s ; rm -fr %s' % (args.d, safe_out), shell=True)
    print(zip_path)



import xarray as xr
import datetime
import numpy as np
import glob
from .utils import bind, url_get
import pandas as pd


def resource_strftime(resource, **kwargs):
    """
    From a resource string like '%Y/%j/myfile_%Y%m%d%H%M.nc' and a date like 'Timestamp('2018-10-13 06:23:22.317102')',
    returns a tuple composed of the closer available date and string like '/2018/286/myfile_201810130600.nc'

    If ressource string is an url (ie 'ftp://ecmwf/%Y/%j/myfile_%Y%m%d%H%M.nc'), fsspec will be used to retreive the file locally.

    Parameters
    ----------
    resource: str

        resource string, with strftime template

    date: datetime

        date to be used

    step: int

        hour step between 2 files


    Returns
    -------
    tuple : (datetime,str)

    """

    date = kwargs["date"]
    step = kwargs["step"]

    delta = datetime.timedelta(hours=step) / 2
    date = date.replace(
        year=(date + delta).year,
        month=(date + delta).month,
        day=(date + delta).day,
        hour=(date + delta).hour // step * step,
        minute=0,
        second=0,
        microsecond=0,
    )

    return date, url_get(date.strftime(resource))


def _to_lon180(ds):
    # roll [0, 360] to [-180, 180] on dim x
    ds = ds.roll(x=-np.searchsorted(ds.x, 180), roll_coords=True)
    ds["x"] = xr.where(ds["x"] >= 180, ds["x"] - 360, ds["x"])
    return ds


def ecmwf_0100_1h(fname, **kwargs):
    """
    ecmwf 0.1 deg 1h reader (ECMWF_FORECAST_0100_202109091300_10U_10V.nc)

    Parameters
    ----------
    fname: str

        hwrf filename

    Returns
    -------
    xarray.Dataset
    """
    ecmwf_ds = xr.open_dataset(
        fname, chunks={"Longitude": 1000, "Latitude": 1000}
    ).isel(time=0)
    ecmwf_ds.attrs["time"] = datetime.datetime.fromtimestamp(
        ecmwf_ds.time.item() // 1000000000
    )
    if "time" in ecmwf_ds:
        ecmwf_ds = ecmwf_ds.drop("time")
    ecmwf_ds = ecmwf_ds[["Longitude", "Latitude", "10U", "10V"]].rename(
        {"Longitude": "x", "Latitude": "y", "10U": "U10", "10V": "V10"}
    )
    ecmwf_ds.attrs = {k: ecmwf_ds.attrs[k] for k in ["title", "institution", "time"]}

    # dataset is lon [0, 360], make it [-180,180]
    ecmwf_ds = _to_lon180(ecmwf_ds)

    ecmwf_ds.rio.write_crs("EPSG:4326", inplace=True)

    return ecmwf_ds


def ecmwf_0125_1h(fname, **kwargs):
    """
    ecmwf 0.125 deg 1h reader (ecmwf_201709071100.nc)

    Parameters
    ----------
    fname: str

        hwrf filename

    Returns
    -------
    xarray.Dataset
    """
    ecmwf_ds = xr.open_dataset(fname, chunks={"longitude": 1000, "latitude": 1000})

    ecmwf_ds = (
        ecmwf_ds.rename({"longitude": "x", "latitude": "y"})
        .rename({"Longitude": "x", "Latitude": "y", "U": "U10", "V": "V10"})
        .set_coords(["x", "y"])
    )

    ecmwf_ds["x"] = ecmwf_ds.x.compute()
    ecmwf_ds["y"] = ecmwf_ds.y.compute()

    # dataset is lon [0, 360], make it [-180,180]
    ecmwf_ds = _to_lon180(ecmwf_ds)

    ecmwf_ds.attrs["time"] = datetime.datetime.fromisoformat(ecmwf_ds.attrs["date"])

    ecmwf_ds.rio.write_crs("EPSG:4326", inplace=True)

    return ecmwf_ds


def hwrf_0015_3h(fname, **kwargs):
    """
    hwrf 0.015 deg 3h reader ()


    Parameters
    ----------
    fname: str

        hwrf filename

    Returns
    -------
    xarray.Dataset
    """
    hwrf_ds = xr.open_dataset(fname)
    try:
        hwrf_ds = hwrf_ds[["U", "V", "LON", "LAT"]]
        hwrf_ds = hwrf_ds.squeeze("t", drop=True)

    except Exception:
        raise ValueError("date '%s' can't be find in %s " % (kwargs["date"], fname))

    hwrf_ds.attrs["time"] = datetime.datetime.strftime(
        kwargs["date"], "%Y-%m-%d %H:%M:%S"
    )

    hwrf_ds = (
        hwrf_ds.assign_coords(
            {"x": hwrf_ds.LON.values[0, :], "y": hwrf_ds.LAT.values[:, 0]}
        )
        .drop_vars(["LON", "LAT"])
        .rename({"U": "U10", "V": "V10"})
    )

    # hwrf_ds.attrs = {k: hwrf_ds.attrs[k] for k in ['institution', 'time']}
    hwrf_ds = _to_lon180(hwrf_ds)
    hwrf_ds.rio.write_crs("EPSG:4326", inplace=True)

    return hwrf_ds


def era5_0250_1h(fname, **kwargs):
    """
    era5 0.250 deg 1h reader ()


    Parameters
    ----------
    fname: str

        era5 filename

    Returns
    -------
    xarray.Dataset
    """

    ds_era5 = xr.open_dataset(fname, chunks={"time": 1})
    ds_era5 = ds_era5[["u10", "v10", "latitude025", "longitude025"]]
    ds_era5 = ds_era5.sel(time=str(kwargs["date"]), method="nearest")
    ds_era5 = ds_era5.drop("time")

    ds_era5 = ds_era5.rename(
        {"longitude025": "x", "latitude025": "y", "u10": "U10", "v10": "V10"}
    )

    ds_era5.attrs["time"] = kwargs["date"]
    ds_era5 = _to_lon180(ds_era5)
    ds_era5.rio.write_crs("EPSG:4326", inplace=True)
    return ds_era5


def gebco(gebco_files):
    """gebco file reader (geotiff from https://www.gebco.net/data_and_products/gridded_bathymetry_data)"""
    return xr.combine_by_coords(
        [
            xr.open_dataset(f, chunks={"x": 1000, "y": 1000})
            .isel(band=0)
            .drop_vars("band")
            for f in gebco_files
        ]
    )


# list available rasters as a pandas dataframe
available_rasters = pd.DataFrame(columns=["resource", "read_function", "get_function"])
available_rasters.loc["gebco"] = [None, gebco, glob.glob]
available_rasters.loc["ecmwf_0100_1h"] = [
    None,
    ecmwf_0100_1h,
    bind(resource_strftime, ..., step=1),
]
available_rasters.loc["ecmwf_0125_1h"] = [
    None,
    ecmwf_0125_1h,
    bind(resource_strftime, ..., step=1),
]
available_rasters.loc["hwrf_0015_3h"] = [
    None,
    hwrf_0015_3h,
    bind(resource_strftime, ..., step=3),
]
available_rasters.loc["era5_0250_1h"] = [
    None,
    era5_0250_1h,
    bind(resource_strftime, ..., step=1),
]



__all__ = [
    "open_dataset",
    "open_datatree",
    "product_info",
    "Sentinel1Meta",
    "Sentinel1Dataset",
    "RadarSat2Dataset",
    "RadarSat2Meta",
    "RcmMeta",
    "RcmDataset",
    "BaseDataset",
    "BaseMeta",
    "get_test_file",
]

from xsar.radarsat2_dataset import RadarSat2Dataset
from xsar.sentinel1_dataset import Sentinel1Dataset
from xsar.sentinel1_meta import Sentinel1Meta
from xsar.rcm_meta import RcmMeta
from xsar.radarsat2_meta import RadarSat2Meta
from xsar.rcm_dataset import RcmDataset
from xsar.base_dataset import BaseDataset
from xsar.base_meta import BaseMeta
from xsar.xsar import open_dataset,open_datatree,product_info
from xsar.xsar import get_test_file

import xsar
try:
    from importlib import metadata
except ImportError: # for Python<3.8
    import importlib_metadata as metadata
try:
    __version__ = metadata.version("xsar")
except Exception:
    __version__ = "999"



import copy
import logging
import warnings

import cartopy
import rasterio
import shapely
from shapely.geometry import Polygon
import numpy as np
from datetime import datetime

from abc import abstractmethod

from .raster_readers import available_rasters
from .base_dataset import BaseDataset
import geopandas as gpd

from .utils import class_or_instancemethod, to_lon180, haversine

logger = logging.getLogger("xsar.base_meta")
logger.addHandler(logging.NullHandler())

# we know tiff as no geotransform : ignore warning
warnings.filterwarnings("ignore", category=rasterio.errors.NotGeoreferencedWarning)

# allow nan without warnings
# some dask warnings are still non filtered: https://github.com/dask/dask/issues/3245
np.errstate(invalid="ignore")


class BaseMeta(BaseDataset):
    """
    Abstract class that defines necessary common functions for the computation of different SAR metadata
    (Radarsat2, Sentinel1, RCM...).
    This also permit a better maintenance, because these functions aren't redefined many times.
    """

    # default mask feature (see self.set_mask_feature and cls.set_mask_feature)
    _mask_features_raw = {
        "land": cartopy.feature.NaturalEarthFeature("physical", "land", "10m")
    }

    _mask_features = {}
    _mask_intersecting_geometries = {}
    _mask_geometry = {}
    _geoloc = None
    _rasterized_masks = None
    manifest_attrs = None
    _time_range = None
    name = None
    multidataset = None
    short_name = None
    path = None
    product = None
    manifest = None
    subdatasets = None
    dsid = None
    safe = None
    geoloc = None

    def __init__(self):
        self.rasters = available_rasters.iloc[0:0].copy()

    def _get_mask_feature(self, name):
        # internal method that returns a cartopy feature from a mask name
        if self._mask_features[name] is None:
            feature = self._mask_features_raw[name]
            if isinstance(feature, str):
                # feature is a shapefile.
                # we get the crs from the shapefile to be able to transform the footprint to this crs_in
                # (so we can use `mask=` in gpd.read_file)
                import fiona
                import pyproj
                from shapely.ops import transform

                with fiona.open(feature) as fshp:
                    try:
                        # proj6 give a " FutureWarning: '+init=<authority>:<code>' syntax is deprecated. "
                        # '<authority>:<code>' is the preferred initialization method"
                        crs_in = fshp.crs["init"]
                    except KeyError:
                        crs_in = fshp.crs
                    crs_in = pyproj.CRS(crs_in)
                proj_transform = pyproj.Transformer.from_crs(
                    pyproj.CRS("EPSG:4326"), crs_in, always_xy=True
                ).transform
                footprint_crs = transform(proj_transform, self.footprint)

                with warnings.catch_warnings():
                    # ignore "RuntimeWarning: Sequential read of iterator was interrupted. Resetting iterator."
                    warnings.simplefilter("ignore", RuntimeWarning)
                    feature = cartopy.feature.ShapelyFeature(
                        gpd.read_file(feature, mask=footprint_crs)
                        .to_crs(epsg=4326)
                        .geometry,
                        cartopy.crs.PlateCarree(),
                    )
            if not isinstance(feature, cartopy.feature.Feature):
                raise TypeError("Expected a cartopy.feature.Feature type")
            self._mask_features[name] = feature

        return self._mask_features[name]

    @class_or_instancemethod
    def set_mask_feature(self_or_cls, name, feature):
        """
        Set a named mask from a shapefile or a cartopy feature.

        Parameters
        ----------
        name: str
            mask name
        feature: str or cartopy.feature.Feature
            if str, feature is a path to a shapefile or whatever file readable with fiona.
            It is recommended to use str, as the serialization of cartopy feature might be big.

        Examples
        --------
            Add an 'ocean' mask at class level (ie as default mask):
            ```
            >>> xsar.RadarSat2Meta.set_mask_feature("ocean", cartopy.feature.OCEAN)
            >>> xsar.Sentinel1Meta.set_mask_feature("ocean", cartopy.feature.OCEAN)
            ```

            Add an 'ocean' mask at instance level (ie only for this self Sentinel1Meta (or RadarSat2Meta instance):
            ```
            >>> xsar.RadarSat2Meta.set_mask_feature("ocean", cartopy.feature.OCEAN)
            >>> xsar.Sentinel1Meta.set_mask_feature("ocean", cartopy.feature.OCEAN)
            ```


            High resoltion shapefiles can be found from openstreetmap.
            It is recommended to use WGS84 with large polygons split from https://osmdata.openstreetmap.de/

        See Also
        --------
        xsar.BaseMeta.get_mask
        """

        # see https://stackoverflow.com/a/28238047/5988771 for self_or_cls

        self_or_cls._mask_features_raw[name] = feature

        if not isinstance(self_or_cls, type):
            # self (instance, not class)
            self_or_cls._mask_intersecting_geometries[name] = None
            self_or_cls._mask_geometry[name] = None
            self_or_cls._mask_features[name] = None

    def get_mask(self, name, describe=False):
        """
        Get mask from `name` (e.g. 'land') as a shapely Polygon.
        The resulting polygon is contained in the footprint.

        Parameters
        ----------
        name: str

        Returns
        -------
        shapely.geometry.Polygon

        """
        if describe:
            descr = self._mask_features_raw[name]
            try:
                # nice repr for a class (like 'cartopy.feature.NaturalEarthFeature land')
                descr = "%s.%s %s" % (
                    descr.__module__,
                    descr.__class__.__name__,
                    descr.name,
                )
            except AttributeError:
                pass
            return descr

        if self._mask_geometry[name] is None:
            if self._get_mask_intersecting_geometries(name).unary_union:
                poly = self._get_mask_intersecting_geometries(
                    name
                ).unary_union.intersection(self.footprint)

            else:
                poly = Polygon()

            if poly.is_empty:
                poly = Polygon()

            self._mask_geometry[name] = poly
        return self._mask_geometry[name]

    def _get_mask_intersecting_geometries(self, name):
        """

        :param name: str(eg land)
        :return:
        """
        if self._mask_intersecting_geometries[name] is None:
            gseries = gpd.GeoSeries(self._get_mask_feature(name).geometries())
            # gseries = gpd.GeoSeries(self._get_mask_feature(name)
            #                         .intersecting_geometries(self.footprint))
            if len(gseries) == 0:
                # no intersection with mask, but we want at least one geometry in the serie (an empty one)
                gseries = gpd.GeoSeries([Polygon()])
            self._mask_intersecting_geometries[name] = gseries
        return self._mask_intersecting_geometries[name]

    @property
    @abstractmethod
    def footprint(self):
        pass

    @property
    def cross_antemeridian(self):
        """True if footprint cross antemeridian"""
        return (
            (np.max(self.geoloc["longitude"]) - np.min(self.geoloc["longitude"])) > 180
        ).item()

    @property
    def swath(self):
        """string like 'EW', 'IW', 'WV', etc ..."""
        return self.manifest_attrs["swath_type"]

    @property
    @abstractmethod
    def _dict_coords2ll(self):
        pass

    @property
    @abstractmethod
    def approx_transform(self):
        pass

    @property
    def mask_names(self):
        """

        Returns
        -------
        list of str
            mask names
        """
        return self._mask_features.keys()

    def coords2ll(self, *args, to_grid=False, approx=False):
        """
        convert `lines`, `samples` arrays to `longitude` and `latitude` arrays.
        or a shapely object in `lines`, `samples` coordinates to `longitude` and `latitude`.

        Parameters
        ----------
        *args: lines, samples  or a shapely geometry
            lines, samples are iterables or scalar

        to_grid: bool, default False
            If True, `lines` and `samples` must be 1D arrays. The results will be 2D array of shape (lines.size, samples.size).

        Returns
        -------
        tuple of np.array or tuple of float
            (longitude, latitude) , with shape depending on `to_grid` keyword.

        See Also
        --------
        xsar.BaseMeta.ll2coords
        xsar.BaseDataset.ll2coords

        """
        if isinstance(args[0], shapely.geometry.base.BaseGeometry):
            return self._coords2ll_shapely(args[0])

        lines, samples = args

        scalar = True
        if hasattr(lines, "__iter__"):
            scalar = False

        if approx:
            if to_grid:
                samples2D, lines2D = np.meshgrid(samples, lines)
                lon, lat = self.approx_transform * (lines2D, samples2D)
                pass
            else:
                lon, lat = self.approx_transform * (lines, samples)
        else:
            dict_coords2ll = self._dict_coords2ll
            if to_grid:
                lon = dict_coords2ll["longitude"](lines, samples)
                lat = dict_coords2ll["latitude"](lines, samples)
            else:
                lon = dict_coords2ll["longitude"].ev(lines, samples)
                lat = dict_coords2ll["latitude"].ev(lines, samples)

        if self.cross_antemeridian:
            lon = to_lon180(lon)

        if scalar and hasattr(lon, "__iter__"):
            lon = lon.item()
            lat = lat.item()

        if hasattr(lon, "__iter__") and type(lon) is not type(lines):
            lon = type(lines)(lon)
            lat = type(lines)(lat)

        return lon, lat

    def _ll2coords_shapely(self, shape, approx=False):
        if approx:
            (xoff, a, b, yoff, d, e) = (~self.approx_transform).to_gdal()
            return shapely.affinity.affine_transform(shape, (a, b, d, e, xoff, yoff))
        else:
            return shapely.ops.transform(self.ll2coords, shape)

    def _coords2ll_shapely(self, shape, approx=False):
        if approx:
            (xoff, a, b, yoff, d, e) = self.approx_transform.to_gdal()
            return shapely.affinity.affine_transform(shape, (a, b, d, e, xoff, yoff))
        else:
            return shapely.ops.transform(self.coords2ll, shape)

    def ll2coords(self, *args):
        """
        Get `(lines, samples)` from `(lon, lat)`,
        or convert a lon/lat shapely object to line/sample coordinates.

        Parameters
        ----------
        *args: lon, lat or shapely object
            lon and lat might be iterables or scalars

        Returns
        -------
        tuple of np.array or tuple of float (lines, samples) , or a shapely object

        Examples
        --------
            get nearest (line,sample) from (lon,lat) = (84.81, 21.32) in ds, without bounds checks

            >>> (line, sample) = self.ll2coords(84.81, 21.32)  # (lon, lat)
            >>> (line, sample)
            (9752.766349989339, 17852.571322887554)

        See Also
        --------
        xsar.BaseMeta.coords2ll
        xsar.BaseDataset.coords2ll

        """

        if isinstance(args[0], shapely.geometry.base.BaseGeometry):
            return self._ll2coords_shapely(args[0])

        lon, lat = args

        # approximation with global inaccurate transform
        line_approx, sample_approx = ~self.approx_transform * (
            np.asarray(lon),
            np.asarray(lat),
        )

        # Theoretical identity. It should be the same, but the difference show the error.
        lon_identity, lat_identity = self.coords2ll(
            line_approx, sample_approx, to_grid=False
        )
        line_identity, sample_identity = ~self.approx_transform * (
            lon_identity,
            lat_identity,
        )

        # we are now able to compute the error, and make a correction
        line_error = line_identity - line_approx
        sample_error = sample_identity - sample_approx

        line = line_approx - line_error
        sample = sample_approx - sample_error

        return line, sample

    def coords2heading(self, lines, samples, to_grid=False, approx=True):
        """
        Get image heading (lines increasing direction) at coords `lines`, `samples`.

        Parameters
        ----------
        lines: np.array or scalar
        samples: np.array or scalar
        to_grid: bool
            If True, `lines` and `samples` must be 1D arrays. The results will be 2D array of shape (lines.size, samples.size).

        Returns
        -------
        np.array or float
            `heading` , with shape depending on `to_grid` keyword.

        """

        lon1, lat1 = self.coords2ll(lines - 1, samples, to_grid=to_grid, approx=approx)
        lon2, lat2 = self.coords2ll(lines + 1, samples, to_grid=to_grid, approx=approx)
        _, heading = haversine(lon1, lat1, lon2, lat2)
        return heading

    @property
    @abstractmethod
    def _get_time_range(self):
        pass

    @property
    def time_range(self):
        """time range as pd.Interval"""
        if self._time_range is None:
            self._time_range = self._get_time_range()
        return self._time_range

    @property
    def start_date(self):
        """start date, as datetime.datetime"""
        out_format = "%Y-%m-%d %H:%M:%S.%f"
        date = self.time_range.left
        try:
            return "%s" % datetime.strptime("%s" % date, out_format)
        except ValueError:
            return "%s" % date.strftime(out_format)

    @property
    def stop_date(self):
        """stop date, as datetime.datetime"""
        out_format = "%Y-%m-%d %H:%M:%S.%f"
        date = self.time_range.right
        try:
            return "%s" % datetime.strptime("%s" % date, out_format)
        except ValueError:
            return "%s" % date.strftime(out_format)

    @class_or_instancemethod
    def set_raster(self_or_cls, name, resource, read_function=None, get_function=None):
        # get defaults if exists
        default = available_rasters.loc[name:name]

        # set from params, or from default
        self_or_cls.rasters.loc[name, "resource"] = (
            resource or default.loc[name, "resource"]
        )
        self_or_cls.rasters.loc[name, "read_function"] = (
            read_function or default.loc[name, "read_function"]
        )
        self_or_cls.rasters.loc[name, "get_function"] = (
            get_function or default.loc[name, "get_function"]
        )

        return

    @property
    def dict(self):
        # return a minimal dictionary that can be used with Sentinel1Meta.from_dict() or pickle (see __reduce__)
        # to reconstruct another instance of self
        #
        minidict = {
            "name": self.name,
            "_mask_features_raw": self._mask_features_raw,
            "_mask_features": {},
            "_mask_intersecting_geometries": {},
            "_mask_geometry": {},
            "rasters": self.rasters,
        }
        for name in minidict["_mask_features_raw"].keys():
            minidict["_mask_intersecting_geometries"][name] = None
            minidict["_mask_geometry"][name] = None
            minidict["_mask_features"][name] = None
        return minidict

    @classmethod
    def from_dict(cls, minidict):
        # like copy constructor, but take a dict from Sentinel1Meta.dict
        # https://github.com/umr-lops/xsar/issues/23
        for name in minidict["_mask_features_raw"].keys():
            assert minidict["_mask_geometry"][name] is None
            assert minidict["_mask_features"][name] is None
        minidict = copy.copy(minidict)
        new = cls(minidict["name"])
        new.__dict__.update(minidict)
        return new



import warnings
from abc import ABC
from datetime import datetime

import dask
import pandas as pd
import shapely
import xarray as xr
import yaml
from affine import Affine
from numpy import asarray
from scipy.interpolate import RectBivariateSpline
from shapely.geometry import Polygon, box
import numpy as np
import logging
import rasterio
from rasterio.control import GroundControlPoint
from shapely.validation import make_valid
import geopandas as gpd
from scipy.spatial import KDTree
import time
import rasterio.features


from xsar.utils import bbox_coords, haversine, map_blocks_coords, timing

logger = logging.getLogger("xsar.base_dataset")
logger.addHandler(logging.NullHandler())

# we know tiff as no geotransform : ignore warning
warnings.filterwarnings("ignore", category=rasterio.errors.NotGeoreferencedWarning)

# allow nan without warnings
# some dask warnings are still non filtered: https://github.com/dask/dask/issues/3245
np.errstate(invalid="ignore")


class BaseDataset(ABC):
    """
    Abstract class that defines necessary common functions for the computation of different SAR dataset variables
    (Radarsat2, Sentinel1, RCM...).
    This also permit a better maintenance, because these functions aren't redefined many times.
    """

    datatree = None
    _dataset = None
    name = None
    sliced = False
    sar_meta = None
    _rasterized_masks = None
    resolution = None
    _da_tmpl = None
    _luts = None
    _map_var_lut = None
    _dtypes = {
        "latitude": "f4",
        "longitude": "f4",
        "incidence": "f4",
        "elevation": "f4",
        "altitude": "f4",
        "ground_heading": "f4",
        "offboresight": "f4",
        "nesz": None,
        "negz": None,
        "sigma0_raw": None,
        "gamma0_raw": None,
        "noise_lut": "f4",
        "noise_lut_range": "f4",
        "noise_lut_azi": "f4",
        "sigma0_lut": "f8",
        "gamma0_lut": "f8",
        "azimuth_time": np.datetime64,
        "slant_range_time": None,
    }
    _default_meta = asarray([], dtype="f8")
    geoloc_tree = None

    @property
    def len_line_m(self):
        """line length, in meters"""
        _bbox_ll = list(zip(*self._bbox_ll))
        len_m, _ = haversine(*_bbox_ll[1], *_bbox_ll[2])
        return len_m

    @property
    def len_sample_m(self):
        """sample length, in meters"""
        _bbox_ll = list(zip(*self._bbox_ll))
        len_m, _ = haversine(*_bbox_ll[0], *_bbox_ll[1])
        return len_m

    @property
    def coverage(self):
        """coverage string"""
        return "%dkm * %dkm (line * sample )" % (
            self.len_line_m / 1000,
            self.len_sample_m / 1000,
        )

    @property
    def _regularly_spaced(self):
        return (
            max(
                [
                    np.unique(np.round(np.diff(self._dataset[dim].values), 1)).size
                    for dim in ["line", "sample"]
                ]
            )
            == 1
        )

    @property
    def _bbox_ll(self):
        """Dataset bounding box, lon/lat"""
        return self.sar_meta.coords2ll(*zip(*self._bbox_coords))

    @property
    def _bbox_coords(self):
        """
        Dataset bounding box, in line/sample coordinates
        """
        bbox_ext = bbox_coords(self.dataset.line.values, self.dataset.sample.values)
        return bbox_ext

    @property
    def geometry(self):
        """
        geometry of this dataset, as a `shapely.geometry.Polygon` (lon/lat coordinates)
        """
        return Polygon(zip(*self._bbox_ll))

    def load_ground_heading(self):
        """
        Load ground heading as delayed thanks to `BaseMeta.coords2heading`.

        Returns
        -------
        xarray.Dataset
            Contains the ground heading
        """

        def coords2heading(lines, samples):
            return self.sar_meta.coords2heading(
                lines, samples, to_grid=True, approx=True
            )

        gh = map_blocks_coords(
            self._da_tmpl.astype(self._dtypes["ground_heading"]),
            coords2heading,
            name="ground_heading",
        )

        gh.attrs = {
            "comment": "at ground level, computed from lon/lat in azimuth direction",
            "long_name": "Platform heading (azimuth from North)",
            "units": "Degrees",
        }

        return gh.to_dataset(name="ground_heading")

    def add_rasterized_masks(self):
        """
        add rasterized masks only (included in add_high_resolution_variables() for Sentinel-1)
        :return:
        """
        self._rasterized_masks = self.load_rasterized_masks()
        # self.datatree['measurement'].ds = xr.merge([self.datatree['measurement'].ds,self._rasterized_masks])
        self.datatree["measurement"] = self.datatree["measurement"].assign(
            xr.merge([self.datatree["measurement"].ds, self._rasterized_masks])
        )

    def recompute_attrs(self):
        """
        Recompute dataset attributes. It's automaticaly called if you assign a new dataset, for example

        >>> xsar_obj.dataset = xsar_obj.dataset.isel(line=slice(1000, 5000))
        >>> # xsar_obj.recompute_attrs() # not needed

        This function must be manually called before using the `.rio` accessor of a variable

        >>> xsar_obj.recompute_attrs()
        >>> xsar_obj.dataset["sigma0"].rio.reproject(...)

        See Also
        --------
            [rioxarray information loss](https://corteva.github.io/rioxarray/stable/getting_started/manage_information_loss.html)

        """
        if not self._regularly_spaced:
            warnings.warn(
                "Irregularly spaced dataset (probably multiple selection). Some attributes will be incorrect."
            )
        attrs = self._dataset.attrs
        # attrs['pixel_sample_m'] = self.pixel_sample_m
        # attrs['pixel_line_m'] = self.pixel_line_m
        attrs["coverage"] = self.coverage
        attrs["footprint"] = self.footprint

        self.dataset.attrs.update(attrs)

        self._dataset = self._set_rio(self._dataset)

        return None

    def coords2ll(self, *args, **kwargs):
        """
        Alias for `xsar.BaseMeta.coords2ll`
        See Also
        --------
        xsar.BaseMeta.coords2ll
        """
        return self.sar_meta.coords2ll(*args, **kwargs)

    def ll2coords(self, *args):
        """
        Get `(lines, samples)` from `(lon, lat)`,
        or convert a lon/lat shapely object to line/sample coordinates.

        Parameters
        ----------
        *args: lon, lat or shapely object

            lon and lat might be iterables or scalars

        Returns
        -------
        tuple of np.array or tuple of float (lines, samples) , or a shapely object

        Notes
        -----
        The difference with `xsar.BaseMeta.ll2coords` is that coordinates are rounded to the nearest dataset coordinates.

        See Also
        --------
        xsar.BaseMeta.ll2coords

        """
        if isinstance(args[0], shapely.geometry.base.BaseGeometry):
            return self.sar_meta._ll2coords_shapely(args[0].intersection(self.geometry))

        line, sample = self.sar_meta.ll2coords(*args)

        if hasattr(args[0], "__iter__"):
            scalar = False
        else:
            scalar = True

        tolerance = (
            np.max(
                [
                    np.percentile(np.diff(self.dataset[c].values), 90) / 2
                    for c in ["line", "sample"]
                ]
            )
            + 1
        )
        try:
            # select the nearest valid pixel in ds
            ds_nearest = self.dataset.sel(
                line=line, sample=sample, method="nearest", tolerance=tolerance
            )
            if scalar:
                (line, sample) = (
                    ds_nearest.line.values.item(),
                    ds_nearest.sample.values.item(),
                )
            else:
                (line, sample) = (ds_nearest.line.values, ds_nearest.sample.values)
        except KeyError:
            # out of bounds, because of `tolerance` keyword
            (line, sample) = (line * np.nan, sample * np.nan)

        return line, sample

    def _set_rio(self, ds):
        # set .rio accessor for ds. ds must be same kind a self._dataset (not checked!)
        gcps = self._local_gcps

        want_dataset = True
        if isinstance(ds, xr.DataArray):
            # temporary convert to dataset
            try:
                ds = ds.to_dataset()
            except ValueError:
                ds = ds.to_dataset(name="_tmp_rio")
            want_dataset = False

        for v in ds:
            if set(["line", "sample"]).issubset(set(ds[v].dims)):
                ds[v] = ds[v].set_index({"sample": "sample", "line": "line"})
                ds[v] = (
                    ds[v]
                    .rio.write_gcps(gcps, "epsg:4326", inplace=True)
                    .rio.set_spatial_dims("sample", "line", inplace=True)
                    .rio.write_coordinate_system(inplace=True)
                )
                # remove/reset some incorrect attrs set by rio
                # (long_name is 'latitude', but it's incorrect for line axis ...)
                for ax in ["line", "sample"]:
                    [
                        ds[v][ax].attrs.pop(k, None)
                        for k in ["long_name", "standard_name"]
                    ]
                    ds[v][ax].attrs["units"] = "1"

        if not want_dataset:
            # convert back to dataarray
            ds = ds[v]
            if ds.name == "_tmp_rio":
                ds.name = None
        return ds

    @property
    def _local_gcps(self):
        # get local gcps, for rioxarray.reproject (row and col are *index*, not coordinates)
        local_gcps = []
        line_decimated = self.dataset.line.values[
            :: int(self.dataset.line.size / 20) + 1
        ]
        sample_decimated = self.dataset.sample.values[
            :: int(self.dataset.sample.size / 20) + 1
        ]
        XX, YY = np.meshgrid(line_decimated, sample_decimated)
        if self.sar_meta.product == "SLC":
            logger.debug(
                "GCPs computed from affine transformations on SLC products can be strongly shifted in position, we advise against ds.rio.reproject()"
            )
        #     lon_s,lat_s = self.coords2ll_SLC(XX.ravel(order='F'),YY.ravel(order='F'))
        #     lon_s = lon_s.values
        #     lat_s = lat_s.values
        cpt = 0
        for line in line_decimated.astype(int):
            for sample in sample_decimated.astype(int):
                irow = np.argmin(np.abs(self.dataset.line.values - line))
                irow = int(irow)
                icol = np.argmin(np.abs(self.dataset.sample.values - sample))
                icol = int(icol)
                # if self.s1meta.product == 'SLC':
                #     #lon, lat = self.coords2ll_SLC(line,sample)
                #     lon = lon_s[cpt]
                #     lat = lat_s[cpt]
                #     if sample%0==0:
                #         print('#########')
                #         print('lon',lon)
                #         print('lat',lat)
                # else:
                lon, lat = self.sar_meta.coords2ll(line, sample)
                gcp = GroundControlPoint(x=lon, y=lat, z=0, col=icol, row=irow)
                local_gcps.append(gcp)
                cpt += 1
        return local_gcps

    def get_burst_valid_location(self):
        """
        add a field 'valid_location' in the bursts sub-group of the datatree

        Returns:
        --------

        """
        nbursts = len(self.datatree["bursts"].ds["burst"])
        burst_firstValidSample = self.datatree["bursts"].ds["firstValidSample"].values
        burst_lastValidSample = self.datatree["bursts"].ds["lastValidSample"].values
        valid_locations = np.empty((nbursts, 4), dtype="int32")
        line_per_burst = len(self.datatree["bursts"].ds["line"])
        for ibur in range(nbursts):
            fvs = burst_firstValidSample[ibur, :]
            lvs = burst_lastValidSample[ibur, :]
            # valind = np.where((fvs != -1) | (lvs != -1))[0]
            valind = np.where(np.isfinite(fvs) | np.isfinite(lvs))[0]
            valloc = [
                ibur * line_per_burst + valind.min(),
                fvs[valind].min(),
                ibur * line_per_burst + valind.max(),
                lvs[valind].max(),
            ]
            valid_locations[ibur, :] = valloc
        tmpda = xr.DataArray(
            dims=["burst", "limits"],
            coords={
                "burst": self.datatree["bursts"].ds["burst"].values,
                "limits": np.arange(4),
            },
            data=valid_locations,
            name="valid_location",
            attrs={
                "description": "start line index, start sample index, stop line index, stop sample index"
            },
        )
        # self.datatree['bursts'].ds['valid_location'] = tmpda
        tmpds = xr.merge([self.datatree["bursts"].ds, tmpda])
        self.datatree["bursts"] = self.datatree["bursts"].assign(tmpds)

    def get_bursts_polygons(self, only_valid_location=True):
        """
        get the polygons of radar bursts in the image geometry

        Parameters
        ----------
        only_valid_location : bool
            [True] -> polygons of the TOPS SLC bursts are cropped using valid location index
            False -> polygons of the TOPS SLC bursts are aligned with azimuth time start/stop index

        Returns
        -------
        geopandas.GeoDataframe
            polygons of the burst in the image (ie line/sample) geometry
            'geometry' is the polygon

        """
        if self.resolution is not None:
            # eg 2.3/100
            factor_range = self.dataset["sampleSpacing"].values / self.resolution
            factor_azimuth = self.dataset["lineSpacing"].values / self.resolution
        else:
            factor_range = 1
            factor_azimuth = 1
        # compute resolution factor if any
        if self.sar_meta.multidataset:
            blocks_list = []
            # for subswath in self.subdatasets.index:
            for submeta in self.sar_meta._submeta:
                block = submeta.bursts(only_valid_location=only_valid_location)
                block["subswath"] = submeta.dsid
                block = block.set_index("subswath", append=True).reorder_levels(
                    ["subswath", "burst"]
                )
                blocks_list.append(block)
            blocks = pd.concat(blocks_list)
        else:
            # burst_list = self._bursts
            self.get_burst_valid_location()
            burst_list = self.datatree["bursts"].ds
            nb_samples = int(
                self.datatree["image"].ds["numberOfSamples"] * factor_range
            )
            if burst_list["burst"].size == 0:
                blocks = gpd.GeoDataFrame()
            else:
                bursts = []
                bursts_az_inds = {}
                inds_burst, geoloc_azitime, geoloc_iburst, geoloc_line = (
                    self._get_indices_bursts()
                )
                for burst_ind, uu in enumerate(np.unique(inds_burst)):
                    if only_valid_location:
                        extent = np.copy(
                            burst_list["valid_location"].values[burst_ind, :]
                        )
                        area = box(
                            int(extent[0] * factor_azimuth),
                            int(extent[1] * factor_range),
                            int(extent[2] * factor_azimuth),
                            int(extent[3] * factor_range),
                        )
                    else:
                        inds_one_val = np.where(inds_burst == uu)[0]
                        bursts_az_inds[uu] = inds_one_val
                        area = box(
                            bursts_az_inds[burst_ind][0],
                            0,
                            bursts_az_inds[burst_ind][-1],
                            nb_samples,
                        )
                    burst = pd.Series(dict([("geometry_image", area)]))
                    bursts.append(burst)
                # to geopandas
                blocks = pd.concat(bursts, axis=1).T
                blocks = gpd.GeoDataFrame(blocks)
                blocks["geometry"] = blocks["geometry_image"].apply(self.coords2ll)
                blocks.index.name = "burst"
        return blocks

    def _get_indices_bursts(self):
        """

        Returns
        -------
        ind np.array
            index of the burst start in the line coordinates
        geoloc_azitime np.array
            azimuth time at the middle of the image from geolocation grid (low resolution)
        geoloc_iburst np.array

        """
        ind = None
        geoloc_azitime = None
        geoloc_iburst = None
        geoloc_line = None
        if self.sar_meta.product == "SLC" and "WV" not in self.sar_meta.swath:
            # if self.datatree.attrs['product'] == 'SLC' and 'WV' not in self.datatree.attrs['swath']:
            burst_nlines = int(self.sar_meta._bursts["linesPerBurst"])
            # burst_nlines = self.datatree['bursts'].ds['line'].size

            geoloc_line = self.sar_meta.geoloc["line"].values
            # geoloc_line = self.datatree['geolocation_annotation'].ds['line'].values
            # find the indice of the bursts in the geolocation grid
            geoloc_iburst = np.floor(geoloc_line / float(burst_nlines)).astype("int32")
            # find the indices of the bursts in the high resolution grid
            line = np.arange(0, self.sar_meta.image["numberOfLines"])
            # line = np.arange(0, self.datatree.attrs['numberOfLines'])
            iburst = np.floor(line / float(burst_nlines)).astype("int32")
            # find the indices of the burst transitions
            ind = np.searchsorted(geoloc_iburst, iburst, side="left")
            n_pixels = int((len(self.sar_meta.geoloc["sample"]) - 1) / 2)
            geoloc_azitime = self.sar_meta.geoloc["azimuthTime"].values[:, n_pixels]
            # security check for unrealistic line_values exceeding the image extent
            if ind.max() >= len(geoloc_azitime):
                ind[ind >= len(geoloc_azitime)] = len(geoloc_azitime) - 1
        return ind, geoloc_azitime, geoloc_iburst, geoloc_line

    @timing
    def load_rasterized_masks(self):
        """
        Load rasterized masks

        Returns
        -------
        xarray.Dataset
            Contains rasterized masks dataset
        """

        def _rasterize_mask_by_chunks(line, sample, mask="land"):
            chunk_coords = bbox_coords(line, sample, pad=None)
            # chunk footprint polygon, in dataset coordinates (with buffer, to enlarge a little the footprint)
            chunk_footprint_coords = Polygon(chunk_coords).buffer(10)
            # chunk footprint polygon, in lon/lat
            chunk_footprint_ll = self.sar_meta.coords2ll(chunk_footprint_coords)

            # get vector mask over chunk, in lon/lat
            vector_mask_ll = self.sar_meta.get_mask(mask).intersection(
                chunk_footprint_ll
            )

            if vector_mask_ll.is_empty:
                # no intersection with mask, return zeros
                return np.zeros((line.size, sample.size))

            # vector mask, in line/sample coordinates
            vector_mask_coords = self.ll2coords(vector_mask_ll)

            # shape of the returned chunk
            out_shape = (line.size, sample.size)

            # transform * (x, y) -> (line, sample)
            # (where (x, y) are index in out_shape)
            # Affine.permutation() is used because (line, sample) is transposed from geographic

            transform = (
                Affine.translation(*chunk_coords[0])
                * Affine.scale(*[np.unique(np.diff(c))[0] for c in [line, sample]])
                * Affine.permutation()
            )

            raster_mask = rasterio.features.rasterize(
                [vector_mask_coords],
                out_shape=out_shape,
                all_touched=False,
                transform=transform,
            )
            return raster_mask

        da_list = []
        for mask in self.sar_meta.mask_names:
            da_mask = map_blocks_coords(
                self._da_tmpl, _rasterize_mask_by_chunks, func_kwargs={"mask": mask}
            )
            name = "%s_mask" % mask
            da_mask.attrs["history"] = yaml.safe_dump(
                {name: self.sar_meta.get_mask(mask, describe=True)}
            )
            da_mask.attrs["meaning"] = "0: ocean , 1: land"
            da_list.append(da_mask.to_dataset(name=name))

        return xr.merge(da_list)

    def ll2coords_SLC(self, *args):
        """
        for SLC product with irregular projected pixel spacing in range Affine transformation are not relevant
        :return:
        """
        stride_dataset_line = 10  # stride !=1 to save computation time, hard coded here to avoid issues of between KDtree serialized and possible different stride usage
        stride_dataset_sample = 30
        # stride_dataset_line = 1 # stride !=1 to save computation time, hard coded here to avoid issues of between KDtree serialized and possible different stride usage
        # stride_dataset_sample = 1
        lon, lat = args
        subset_lon = self.dataset["longitude"].isel(
            {
                "line": slice(None, None, stride_dataset_line),
                "sample": slice(None, None, stride_dataset_sample),
            }
        )
        subset_lat = self.dataset["latitude"].isel(
            {
                "line": slice(None, None, stride_dataset_line),
                "sample": slice(None, None, stride_dataset_sample),
            }
        )
        if self.geoloc_tree is None:
            t0 = time.time()
            lontmp = subset_lon.values.ravel()
            lattmp = subset_lat.values.ravel()
            self.geoloc_tree = KDTree(np.c_[lontmp, lattmp])
            logger.debug("tree ready in %1.2f sec" % (time.time() - t0))
        ll = np.vstack([lon, lat]).T
        dd, ii = self.geoloc_tree.query(ll, k=1)
        line, sample = np.unravel_index(ii, subset_lat.shape)
        return line * stride_dataset_line, sample * stride_dataset_sample

    def coords2ll_SLC(self, *args):
        """
            for SLC product with irregular projected pixel spacing in range Affine transformation are not relevant

        Returns
        -------
        """
        lines, samples = args
        if isinstance(lines, list) or isinstance(lines, np.ndarray):
            pass
        else:  # in case of a single point,
            # to avoid error when declaring the da_line and da_sample below
            lines = [lines]
            samples = [samples]
        da_line = xr.DataArray(lines, dims="points")
        da_sample = xr.DataArray(samples, dims="points")
        lon = self.dataset["longitude"].sel(line=da_line, sample=da_sample)
        lat = self.dataset["latitude"].sel(line=da_line, sample=da_sample)
        return lon, lat

    def land_mask_slc_per_bursts(self, lazy_loading=True):
        """
        1) loop on burst polygons to get rasterized landmask
        2) merge the landmask pieces into a single Dataset to replace existing 'land_mask' is any

        Parameters
        ----------

        lazy_loading bool

        Returns
        -------

        """
        # TODO: add a prior step to compute the intersection between the self.dataset (could be a subset) and the different bursts
        # if 'land_mask' in self.dataset:
        #     self.datatree['measurement'] = self.datatree['measurement'].assign(self.datatree['measurement'].to_dataset().drop('land_mask'))
        logger.debug("start land_mask_slc_per_bursts()")

        def _rasterize_mask_by_chunks(line, sample, mask="land"):
            """
            copy/pasted from load_rasterized_masks()
            :param line:
            :param sample:
            :param mask:
            :return:
            """
            chunk_coords = bbox_coords(line, sample, pad=None)
            # chunk footprint polygon, in dataset coordinates (with buffer, to enlarge a little the footprint)
            # .buffer(1) #no buffer-> corruption of the polygon
            chunk_footprint_coords = Polygon(chunk_coords)
            assert chunk_footprint_coords.is_valid
            lines, samples = chunk_footprint_coords.exterior.xy
            lines = np.array([hh for hh in lines]).astype(int)
            lines = np.clip(lines, a_min=0, a_max=self.dataset["line"].max().values)
            samples = np.array([hh for hh in samples]).astype(int)
            samples = np.clip(
                samples, a_min=0, a_max=self.dataset["sample"].max().values
            )
            chunk_footprint_lon, chunk_footprint_lat = self.coords2ll_SLC(
                lines, samples
            )
            chunk_footprint_ll = Polygon(
                np.vstack([chunk_footprint_lon, chunk_footprint_lat]).T
            )
            if chunk_footprint_ll.is_valid is False:
                chunk_footprint_ll = make_valid(chunk_footprint_ll)
            # get vector mask over chunk, in lon/lat
            vector_mask_ll = self.sar_meta.get_mask(mask).intersection(
                chunk_footprint_ll
            )
            if vector_mask_ll.is_empty:
                # no intersection with mask, return zeros
                return np.zeros((line.size, sample.size))

            # vector mask, in line/sample coordinates
            if isinstance(vector_mask_ll, shapely.geometry.Polygon):
                lons_ma, lats_ma = vector_mask_ll.exterior.xy
                lons = np.array([hh for hh in lons_ma])
                lats = np.array([hh for hh in lats_ma])
                vector_mask_coords_lines, vector_mask_coords_samples = (
                    self.ll2coords_SLC(lons, lats)
                )
                vector_mask_coords = [
                    Polygon(
                        np.vstack(
                            [vector_mask_coords_lines, vector_mask_coords_samples]
                        ).T
                    )
                ]
            else:  # multipolygon
                vector_mask_coords = []  # to store polygons in image coordinates
                for iio, onepoly in enumerate(vector_mask_ll.geoms):
                    lons_ma, lats_ma = onepoly.exterior.xy
                    lons = np.array([hh for hh in lons_ma])
                    lats = np.array([hh for hh in lats_ma])
                    vector_mask_coords_lines, vector_mask_coords_samples = (
                        self.ll2coords_SLC(lons, lats)
                    )
                    vector_mask_coords.append(
                        Polygon(
                            np.vstack(
                                [vector_mask_coords_lines, vector_mask_coords_samples]
                            ).T
                        )
                    )

            # shape of the returned chunk
            out_shape = (line.size, sample.size)

            # transform * (x, y) -> (line, sample)
            # (where (x, y) are index in out_shape)
            # Affine.permutation() is used because (line, sample) is transposed from geographic
            # this transform Affine seems to be sufficient approx for SLC -> curvilinear could be even better?
            transform = (
                Affine.translation(*chunk_coords[0])
                * Affine.scale(*[np.unique(np.diff(c))[0] for c in [line, sample]])
                * Affine.permutation()
            )

            raster_mask = rasterio.features.rasterize(
                vector_mask_coords,
                out_shape=out_shape,
                all_touched=False,
                transform=transform,
            )

            return raster_mask

        # all_bursts = self.datatree['bursts']
        all_bursts = self.get_bursts_polygons(only_valid_location=False)
        da_dict = {}  # dictionnary to store the DataArray of each mask and each burst
        for burst_id in range(len(all_bursts["geometry_image"])):
            a_burst_bbox = all_bursts["geometry_image"].iloc[burst_id]
            line_index = np.array([int(jj) for jj in a_burst_bbox.exterior.xy[0]])
            sample_index = np.array([int(jj) for jj in a_burst_bbox.exterior.xy[1]])
            logger.debug(
                "line_index : %s %s %s", line_index, line_index.min(), line_index.max()
            )
            logger.debug("dataset shape %s", self.dataset.digital_number.shape)
            a_burst_subset = self.dataset.isel(
                {
                    "line": slice(line_index.min(), line_index.max()),
                    "sample": slice(sample_index.min(), sample_index.max()),
                    "pol": 0,
                }
            )
            logger.debug("a_burst_subset %s", a_burst_subset.digital_number.shape)
            # logging.info('burst : %s lines: %s samples: %s',burst_id,a_burst_subset.digital_number.line,a_burst_subset.digital_number.sample)
            da_tmpl = xr.DataArray(
                dask.array.empty_like(
                    np.empty(
                        (
                            len(a_burst_subset.digital_number.line),
                            len(a_burst_subset.digital_number.sample),
                        )
                    ),
                    dtype=np.int8,
                    name="empty_var_tmpl-%s" % dask.base.tokenize(self.sar_meta.name),
                ),
                dims=("line", "sample"),
                coords={
                    "line": a_burst_subset.digital_number.line,
                    "sample": a_burst_subset.digital_number.sample,
                    # 'line_time': line_time.astype(float),
                },
            )

            for mask in self.sar_meta.mask_names:
                logger.debug("mask: %s", mask)
                if lazy_loading:
                    da_mask = map_blocks_coords(
                        da_tmpl, _rasterize_mask_by_chunks, func_kwargs={"mask": mask}
                    )
                else:
                    tmpmask_val = _rasterize_mask_by_chunks(
                        a_burst_subset.digital_number.line,
                        a_burst_subset.digital_number.sample,
                        mask=mask,
                    )
                    da_mask = xr.DataArray(
                        tmpmask_val,
                        dims=("line", "sample"),
                        coords={
                            "line": a_burst_subset.digital_number.line,
                            "sample": a_burst_subset.digital_number.sample,
                        },
                    )
                name = "%s_maskv2" % mask
                da_mask.attrs["history"] = yaml.safe_dump(
                    {name: self.sar_meta.get_mask(mask, describe=True)}
                )
                da_mask.attrs["meaning"] = "0: ocean , 1: land"
                da_mask = da_mask.fillna(0)  # zero -> ocean
                da_mask = da_mask.astype(np.int8)
                logger.debug("%s -> %s", mask, da_mask.attrs["history"])
                # da_list.append(da_mask.to_dataset(name=name))
                if mask not in da_dict:
                    da_dict[mask] = [da_mask.to_dataset(name=name)]
                else:
                    da_dict[mask].append(da_mask.to_dataset(name=name))
            logger.debug("da_dict[mask] = %s %s", mask, da_dict[mask])
        # merge with existing dataset
        all_masks = []
        for kk in da_dict:
            logger.debug("da_dict[kk] %s", da_dict[kk])
            complet_mask = xr.combine_by_coords(da_dict[kk])
            all_masks.append(complet_mask)
        tmpds = self.datatree["measurement"].to_dataset()
        tmpds.attrs = self.dataset.attrs
        tmpmerged = xr.merge([tmpds] + all_masks)
        tmpmerged = tmpmerged.drop("land_mask")
        logger.debug("rename land_maskv2 -> land_mask")
        tmpmerged = tmpmerged.rename({"land_maskv2": "land_mask"})
        tmpmerged.attrs["land_mask_computed_by_burst"] = True
        self.dataset = tmpmerged
        self.datatree["measurement"] = self.datatree["measurement"].assign(tmpmerged)
        self.datatree["measurement"].attrs = tmpmerged.attrs

    @property
    def footprint(self):
        """alias for `xsar.BaseDataset.geometry`"""
        return self.geometry

    @timing
    def map_raster(self, raster_ds):
        """
        Map a raster onto xsar grid

        Parameters
        ----------
        raster_ds: xarray.Dataset or xarray.DataArray
            The dataset we want to project onto xsar grid. The `raster_ds.rio` accessor must be valid.

        Returns
        -------
        xarray.Dataset or xarray.DataArray
            The projected dataset, with 'line' and 'sample' coordinate (same size as xsar dataset), and with valid `.rio` accessor.


        """
        if not raster_ds.rio.crs.is_geographic:
            raster_ds = raster_ds.rio.reproject(4326)

        if self.sar_meta.cross_antemeridian:
            raise NotImplementedError("Antimeridian crossing not yet checked")

        # get lon/lat box for xsar dataset
        lon1, lat1, lon2, lat2 = self.sar_meta.footprint.exterior.bounds
        lon_range = [lon1, lon2]
        lat_range = [lat1, lat2]

        # ensure dims ordering
        raster_ds = raster_ds.transpose("y", "x")

        # ensure coords are increasing ( for RectBiVariateSpline )
        for coord in ["x", "y"]:
            if raster_ds[coord].values[-1] < raster_ds[coord].values[0]:
                raster_ds = raster_ds.reindex({coord: raster_ds[coord][::-1]})

        # from lon/lat box in xsar dataset, get the corresponding box in raster_ds (by index)
        """
        ilon_range = [
            np.searchsorted(raster_ds.x.values, lon_range[0]),
            np.searchsorted(raster_ds.x.values, lon_range[1])
        ]
        ilat_range = [
            np.searchsorted(raster_ds.y.values, lat_range[0]),
            np.searchsorted(raster_ds.y.values, lat_range[1])
        ]
        """  # for incomplete raster (not global like hwrf)
        ilon_range = [
            np.max([1, np.searchsorted(raster_ds.x.values, lon_range[0])]),
            np.min(
                [np.searchsorted(raster_ds.x.values, lon_range[1]), raster_ds.x.size]
            ),
        ]
        ilat_range = [
            np.max([1, np.searchsorted(raster_ds.y.values, lat_range[0])]),
            np.min(
                [np.searchsorted(raster_ds.y.values, lat_range[1]), raster_ds.y.size]
            ),
        ]

        # enlarge the raster selection range, for correct interpolation
        ilon_range, ilat_range = [
            [rg[0] - 1, rg[1] + 1] for rg in (ilon_range, ilat_range)
        ]

        # select the xsar box in the raster
        raster_ds = raster_ds.isel(x=slice(*ilon_range), y=slice(*ilat_range))

        # upscale coordinates, in original projection
        # 1D array of lons/lats, trying to have same spacing as dataset (if not to high)
        num = min((self._dataset.sample.size + self._dataset.line.size) // 2, 1000)
        lons = np.linspace(*lon_range, num=num)
        lats = np.linspace(*lat_range, num=num)

        name = None
        if isinstance(raster_ds, xr.DataArray):
            # convert to temporary dataset
            name = raster_ds.name or "_tmp_name"
            raster_ds = raster_ds.to_dataset(name=name)

        mapped_ds_list = []
        for var in raster_ds:
            raster_da = raster_ds[var].chunk(raster_ds[var].shape)
            # upscale in original projection using interpolation
            # in most cases, RectBiVariateSpline give better results, but can't handle Nans
            if np.any(np.isnan(raster_da)):
                upscaled_da = raster_da.interp(x=lons, y=lats)
            else:
                upscaled_da = map_blocks_coords(
                    xr.DataArray(dims=["y", "x"], coords={"x": lons, "y": lats}).chunk(
                        1000
                    ),
                    RectBivariateSpline(
                        raster_da.y.values,
                        raster_da.x.values,
                        raster_da.values,
                        kx=3,
                        ky=3,
                    ),
                )
            upscaled_da.name = var
            # interp upscaled_da on sar grid
            mapped_ds_list.append(
                upscaled_da.interp(
                    x=self._dataset.longitude, y=self._dataset.latitude
                ).drop_vars(["x", "y"])
            )
        mapped_ds = xr.merge(mapped_ds_list)

        if name is not None:
            # convert back to dataArray
            mapped_ds = mapped_ds[name]
            if name == "_tmp_name":
                mapped_ds.name = None
        return self._set_rio(mapped_ds)

    @timing
    def _load_rasters_vars(self):
        # load and map variables from rasterfile (like ecmwf) on dataset
        if self.sar_meta.rasters.empty:
            return None
        else:
            logger.warning("Raster variable are experimental")

        if self.sar_meta.cross_antemeridian:
            raise NotImplementedError("Antimeridian crossing not yet checked")

        # will contain xr.DataArray to merge
        da_var_list = []

        for name, infos in self.sar_meta.rasters.iterrows():
            # read the raster file using helpers functions
            read_function = infos["read_function"]
            get_function = infos["get_function"]
            resource = infos["resource"]

            kwargs_get = {
                "date": datetime.strptime(
                    self.sar_meta.start_date, "%Y-%m-%d %H:%M:%S.%f"
                ),
                "footprint": self.sar_meta.footprint,
            }

            logger.debug(
                'adding raster "%s" from resource "%s"' % (name, str(resource))
            )
            if get_function is not None:
                try:
                    resource_dec = get_function(resource, **kwargs_get)
                except TypeError:
                    resource_dec = get_function(resource)

            kwargs_read = {"date": np.datetime64(resource_dec[0])}

            if read_function is None:
                raster_ds = xr.open_dataset(resource_dec[1], chunk=1000)
            else:
                # read_function should return a chunked dataset (so it's fast)
                raster_ds = read_function(resource_dec[1], **kwargs_read)

            # add globals raster attrs to globals dataset attrs
            hist_res = {"resource": resource_dec[1]}
            if get_function is not None:
                hist_res.update({"resource_decoded": resource_dec[1]})

            reprojected_ds = self.map_raster(raster_ds).rename(
                {v: "%s_%s" % (name, v) for v in raster_ds}
            )

            for v in reprojected_ds:
                reprojected_ds[v].attrs["history"] = yaml.safe_dump({v: hist_res})

            da_var_list.append(reprojected_ds)
        return xr.merge(da_var_list)

    def _get_lut(self, var_name):
        """
        Get lut for `var_name`

        Parameters
        ----------
        var_name: str

        Returns
        -------
        xarray.DataArray
            lut for `var_name`
        """
        try:
            lut_name = self._map_var_lut[var_name]
        except KeyError:
            raise ValueError("can't find lut name for var '%s'" % var_name)
        try:
            lut = self._luts[lut_name]
        except KeyError:
            raise ValueError(
                "can't find lut from name '%s' for variable '%s' "
                % (lut_name, var_name)
            )
        return lut



try:
    # will fall back to repr if some modules are missing

    # make sure we are running from a notebook
    # if test fail, nothing will be imported, and that will save lot of importtime
    assert get_ipython() is not None
    import cartopy
    import holoviews as hv
    import geoviews as gv
    import geoviews.feature as gf
    import jinja2
    import geopandas as gpd
    import holoviews.ipython.display_hooks as display_hooks
    from shapely.geometry import Polygon
except (ModuleNotFoundError, AssertionError, NameError):
    pass


def repr_mimebundle_Sentinel1Meta(self, include=None, exclude=None):
    """html output for notebook"""

    template = jinja2.Template(
        """
        <div align="left">
            <h5>{{ intro }}</h5>
            <table style="width:100%">
                <thead>
                    <tr>
                        <th colspan="2">{{ short_name }}</th>
                    </tr>
                </thead>
                <tbody>
                    <tr>
                        <td>
                            <table>
                                {% for key, value in properties.items() %}
                                 <tr>
                                     <th> {{ key }} </th>
                                     <td> {{ value }} </td>
                                 </tr>
                                {% endfor %}
                            </table>
                        </td>
                        <td>{{ location }}</td>
                    </tr>
                </tbody>
            </table>

        </div>

        """
    )

    crs = cartopy.crs.PlateCarree()

    world = gv.operation.resample_geometry(gf.land.geoms("10m")).opts(
        color="khaki", projection=crs, alpha=0.5
    )

    center = self.footprint.centroid
    xlim = (center.x - 20, center.x + 20)
    ylim = (center.y - 20, center.y + 20)

    if self.multidataset and len(self.subdatasets) == len(
        self._footprints
    ):  # checks len because SAFEs like IW_SLC has only one footprint for 3 subdatasets
        dsid = [s.split(":")[2] for s in self.subdatasets]
        footprint = self._footprints
    else:
        dsid = [self.dsid]
        footprint = [self.footprint]

    footprints_df = gpd.GeoDataFrame(
        {"dsid": dsid, "geometry": footprint}, crs="EPSG:4326"
    )

    opts = {"bokeh": dict(tools=["hover"])}

    footprint = (
        gv.Polygons(footprints_df, label="footprint")
        .opts(projection=crs, xlim=xlim, ylim=ylim, alpha=0.5)
        .opts(
            **(opts.get(hv.Store.current_backend) or {}),
            backend=hv.Store.current_backend
        )
    )

    orbit = (
        gv.Points(self.orbit["geometry"].to_crs("EPSG:4326"), label="orbit")
        .opts(projection=crs, xlim=xlim, ylim=ylim, alpha=0.5)
        .opts(
            **(opts.get(hv.Store.current_backend) or {}),
            backend=hv.Store.current_backend
        )
    )

    opts = {"bokeh": dict(width=400, height=400), "matplotlib": dict(fig_inches=5)}
    location = (
        (world * footprint * orbit)
        .opts(title="Map")
        .opts(
            **(opts.get(hv.Store.current_backend) or {}),
            backend=hv.Store.current_backend
        )
    )
    data, metadata = display_hooks.render(location)

    properties = self.to_dict()
    properties["orbit_pass"] = self.orbit_pass
    if self.pixel_line_m is not None:
        properties["pixel size"] = "%.1f * %.1f meters (line * sample)" % (
            self.pixel_line_m,
            self.pixel_sample_m,
        )
    properties["coverage"] = self.coverage
    properties["start_date"] = self.start_date
    properties["stop_date"] = self.stop_date
    if len(self.subdatasets) > 0:
        properties["subdatasets"] = "list of %d subdatasets" % len(self.subdatasets)
    properties = {k: v for k, v in properties.items() if v is not None}

    if self.multidataset:
        intro = "Multi (%d) dataset" % len(self.subdatasets)
    else:
        intro = "Single dataset"
        properties["dsid"] = self.dsid

    if "text/html" in data:
        data["text/html"] = template.render(
            intro=intro,
            short_name=self.short_name,
            properties=properties,
            location=data["text/html"],
        )

    return data, metadata


def repr_mimebundle_Sentinel1Dataset(self, include=None, exclude=None):

    template = jinja2.Template(
        """
        <div align="left">
            <h5>{{ intro }}</h5>
            <table style="width:100%">
                <thead>
                    <tr>
                        <th colspan="2">{{ short_name }}</th>
                    </tr>
                </thead>
                <tbody>
                    <tr>
                        <td>
                            <table>
                                {% for key, value in properties.items() %}
                                 <tr>
                                     <th> {{ key }} </th>
                                     <td> {{ value }} </td>
                                 </tr>
                                {% endfor %}
                            </table>
                        </td>
                        <td>{{ location }}</td>
                    </tr>
                </tbody>
            </table>

        </div>

        """
    )

    opts = {"bokeh": dict(fill_color="cyan")}
    grid = (
        hv.Path(Polygon(self._bbox_coords_ori)).opts(color="blue")
        * hv.Polygons(Polygon(self._bbox_coords))
        .opts(color="blue")
        .opts(
            **(opts.get(hv.Store.current_backend) or {}),
            backend=hv.Store.current_backend
        )
    ).opts(xlabel="line", ylabel="sample")

    data, metadata = display_hooks.render(grid)
    properties = {}
    if self.pixel_line_m is not None:
        properties["pixel size"] = "%.1f * %.1f meters (line * sample)" % (
            self.pixel_line_m,
            self.pixel_sample_m,
        )
    properties["coverage"] = self.coverage
    properties = {k: v for k, v in properties.items() if v is not None}

    if self.sliced:
        intro = "dataset slice"
    else:
        intro = "full dataset coverage"

    if "text/html" in data:
        data["text/html"] = template.render(
            intro=intro,
            short_name=self.sar_meta.short_name,
            properties=properties,
            location=data["text/html"],
        )

    return data, metadata


repr_mimebundle_wrapper = {
    "Sentinel1Meta": repr_mimebundle_Sentinel1Meta,
    "Sentinel1Dataset": repr_mimebundle_Sentinel1Dataset,
}


def repr_mimebundle(obj, include=None, exclude=None):
    try:
        return repr_mimebundle_wrapper[type(obj).__name__](
            obj, include=include, exclude=exclude
        )
    except Exception:
        return {"text": repr(obj)}



# -*- coding: utf-8 -*-
import logging
import numpy as np
import xarray
import pandas as pd
import geopandas as gpd
import rasterio
from rasterio.control import GroundControlPoint
from scipy.interpolate import RectBivariateSpline
from shapely.geometry import Polygon
from shapely.ops import unary_union

from .base_meta import BaseMeta
from .utils import haversine, timing
import os
from .ipython_backends import repr_mimebundle


logger = logging.getLogger("xsar.sentinel1_meta")
logger.addHandler(logging.NullHandler())


class Sentinel1Meta(BaseMeta):
    """
    Handle dataset metadata.
    A `xsar.Sentinel1Meta` object can be used with `xsar.open_dataset`,
    but it can be used as itself: it contains usefull attributes and methods.

    Parameters
    ----------
    name: str
        path or gdal identifier like `'SENTINEL1_DS:%s:WV_001' % path`

    """

    # class attributes are needed to fetch instance attribute (ie self.name) with dask actors
    # ref http://distributed.dask.org/en/stable/actors.html#access-attributes
    # FIXME: not needed if @property, so it might be a good thing to have getter for those attributes
    xsd_definitions = None

    @timing
    def __init__(self, name):
        super().__init__()

        try:
            from safe_s1.metadata import Sentinel1Reader
        except ImportError:
            from safe_s1.reader import Sentinel1Reader
        self.reader = Sentinel1Reader(name)

        if not name.startswith("SENTINEL1_DS:"):
            name = name.rstrip("/")  # remove trailing space
            name = "SENTINEL1_DS:%s:" % name
        else:
            name = name.replace("/:", ":")
        self.name = name
        """Gdal dataset name"""
        name_parts = self.name.split(":")
        if len(name_parts) > 3:
            # windows might have semicolon in path ('c:\...')
            name_parts[1] = ":".join(name_parts[1:-1])
            del name_parts[2:-1]
        name_parts[1] = os.path.basename(name_parts[1])
        self.short_name = ":".join(name_parts)
        """Like name, but without path"""
        self.path = ":".join(self.name.split(":")[1:-1])
        """Dataset path"""
        self.safe = os.path.basename(self.path)
        """Safe file name"""
        # there is no information on resolution 'F' 'H' or 'M' in the manifest, so we have to extract it from filename
        try:
            self.product = os.path.basename(self.path).split("_")[2]
        except ValueError:
            print("path: %s" % self.path)
            self.product = "XXX"
        """Product type, like 'GRDH', 'SLC', etc .."""

        self.dt = self.reader.datatree

        self.manifest_attrs = self.reader.manifest_attrs

        self.multidataset = False
        """True if multi dataset"""
        self.subdatasets = gpd.GeoDataFrame(geometry=[], index=[])
        """Subdatasets as GeodataFrame (empty if single dataset)"""
        datasets_names = self.reader.datasets_names
        self.xsd_definitions = self.reader.get_annotation_definitions()
        if self.name.endswith(":") and len(datasets_names) == 1:
            self.name = datasets_names[0]
        self.dsid = self.name.split(":")[-1]
        """Dataset identifier (like 'WV_001', 'IW1', 'IW'), or empty string for multidataset"""
        # submeta is a list of submeta objects if multidataset and TOPS
        # this list will remain empty for _WV__SLC because it will be time-consuming to process them
        self._submeta = []
        if self.short_name.endswith(":"):
            self.short_name = self.short_name + self.dsid
        if self.reader.files.empty:
            try:
                self.subdatasets = gpd.GeoDataFrame(
                    geometry=self.manifest_attrs["footprints"], index=datasets_names
                )
            except ValueError:
                # not as many footprints than subdatasets count. (probably TOPS product)
                self._submeta = [Sentinel1Meta(subds) for subds in datasets_names]
                sub_footprints = [submeta.footprint for submeta in self._submeta]
                self.subdatasets = gpd.GeoDataFrame(
                    geometry=sub_footprints, index=datasets_names
                )
            self.multidataset = True

        self.platform = (
            self.manifest_attrs["mission"] + self.manifest_attrs["satellite"]
        )
        """Mission platform"""
        self._time_range = None
        for name, feature in self.__class__._mask_features_raw.items():
            self.set_mask_feature(name, feature)
        """pandas dataframe for rasters (see `xsar.Sentinel1Meta.set_raster`)"""

    def __del__(self):
        logger.debug("__del__")

    def have_child(self, name):
        """
        Check if dataset `name` belong to this Sentinel1Meta object.

        Parameters
        ----------
        name: str
            dataset name

        Returns
        -------
        bool
        """
        return name == self.name or name in self.subdatasets.index

    def _get_time_range(self):
        if self.multidataset:
            time_range = [
                self.manifest_attrs["start_date"],
                self.manifest_attrs["stop_date"],
            ]
        else:
            time_range = self.reader.time_range
        return pd.Interval(
            left=pd.Timestamp(time_range[0]),
            right=pd.Timestamp(time_range[-1]),
            closed="both",
        )

    def to_dict(self, keys="minimal"):

        info_keys = {"minimal": ["ipf", "platform", "swath", "product", "pols"]}
        info_keys["all"] = info_keys["minimal"] + [
            "name",
            "start_date",
            "stop_date",
            "footprint",
            "coverage",
            "orbit_pass",
            "platform_heading",
        ]  # 'pixel_line_m', 'pixel_sample_m',

        if isinstance(keys, str):
            keys = info_keys[keys]

        res_dict = {}
        for k in keys:
            if hasattr(self, k):
                res_dict[k] = getattr(self, k)
            elif k in self.manifest_attrs.keys():
                res_dict[k] = self.manifest_attrs[k]
            else:
                raise KeyError('Unable to find key/attr "%s" in Sentinel1Meta' % k)
        return res_dict

    def annotation_angle(self, line, sample, angle):
        """Interpolate angle with RectBivariateSpline"""
        lut = angle.reshape(line.size, sample.size)
        lut_f = RectBivariateSpline(line, sample, lut, kx=1, ky=1)
        return lut_f

    @property
    def orbit_pass(self):
        """
        Orbit pass, i.e 'Ascending' or 'Descending'
        """

        if self.multidataset:
            return None  # not defined for multidataset

        return self.orbit.attrs["orbit_pass"]

    @property
    def platform_heading(self):
        """
        Platform heading, relative to north
        """

        if self.multidataset:
            return None  # not defined for multidataset

        return self.orbit.attrs["platform_heading"]

    @property
    def rio(self):
        raise DeprecationWarning(
            "Sentinel1Meta.rio is deprecated. "
            'Use `rasterio.open` on files in `Sentinel1Meta..files["measurement"] instead`'
        )

    @property
    def footprint(self):
        """footprint, as a shapely polygon or multi polygon"""
        if self.multidataset:
            return unary_union(self._footprints)
        return self.geoloc.attrs["footprint"]

    @property
    def geometry(self):
        """alias for footprint"""
        return self.footprint

    @property
    def geoloc(self):
        """
        xarray.Dataset with `['longitude', 'latitude', 'altitude', 'azimuth_time', 'slant_range_time','incidence','elevation' ]` variables
        and `['line', 'sample']` coordinates, at the geolocation grid
        """
        if self.multidataset:
            raise TypeError("geolocation_grid not available for multidataset")
        if self._geoloc is None:
            self._geoloc = self.dt["geolocationGrid"].to_dataset()
            self._geoloc.attrs = {}
            # compute attributes (footprint, coverage, pixel_size)
            footprint_dict = {}
            for ll in ["longitude", "latitude"]:
                footprint_dict[ll] = [
                    self._geoloc[ll].isel(line=a, sample=x).values
                    for a, x in [(0, 0), (0, -1), (-1, -1), (-1, 0)]
                ]
            corners = list(zip(footprint_dict["longitude"], footprint_dict["latitude"]))
            p = Polygon(corners)
            self._geoloc.attrs["footprint"] = p

            # compute acquisition size/resolution in meters
            # first vector is on sample
            acq_sample_meters, _ = haversine(*corners[0], *corners[1])
            # second vector is on line
            acq_line_meters, _ = haversine(*corners[1], *corners[2])
            self._geoloc.attrs["coverage"] = "%dkm * %dkm (line * sample )" % (
                acq_line_meters / 1000,
                acq_sample_meters / 1000,
            )

            # compute self._geoloc.attrs['approx_transform'], from gcps
            # we need to convert self._geoloc to  a list of GroundControlPoint
            def _to_rio_gcp(pt_geoloc):
                # convert a point from self._geoloc grid to rasterio GroundControlPoint
                return GroundControlPoint(
                    x=pt_geoloc.longitude.item(),
                    y=pt_geoloc.latitude.item(),
                    z=pt_geoloc.height.item(),
                    col=pt_geoloc.line.item(),
                    row=pt_geoloc.sample.item(),
                )

            gcps = [
                _to_rio_gcp(self._geoloc.sel(line=line, sample=sample))
                for line in self._geoloc.line
                for sample in self._geoloc.sample
            ]
            # approx transform, from all gcps (inaccurate)
            self._geoloc.attrs["approx_transform"] = rasterio.transform.from_gcps(gcps)
            for vv in self._geoloc:
                if vv in self.xsd_definitions:
                    self._geoloc[vv].attrs["definition"] = str(self.xsd_definitions[vv])

        return self._geoloc

    @property
    def _footprints(self):
        """footprints as list. should len 1 for single meta, or len(self.subdatasets) for multi meta"""
        return self.manifest_attrs["footprints"]

    @property
    def coverage(self):
        """coverage, as a string like '251km * 170km (sample * line )'"""
        if self.multidataset:
            return None  # not defined for multidataset
        return self.geoloc.attrs["coverage"]

    @property
    def pixel_line_m(self):
        """pixel line spacing, in meters (at sensor level)"""
        if self.multidataset:
            res = None  # not defined for multidataset
        else:
            res = self.image["azimuthPixelSpacing"]
        return res

    @property
    def pixel_sample_m(self):
        """pixel sample spacing, in meters (at sensor level)"""
        if self.multidataset:
            res = None  # not defined for multidataset
        else:
            res = self.image["groundRangePixelSpacing"]
        return res

    @property
    def denoised(self):
        """dict with pol as key, and bool as values (True is DN is predenoised at L1 level)"""
        return self.reader.denoised

    @property
    def ipf(self):
        """ipf version"""
        return self.manifest_attrs["ipf_version"]

    @property
    def pols(self):
        """polarisations strings, separated by spaces"""
        return " ".join(self.manifest_attrs["polarizations"])

    @property
    def orbit(self):
        """
        orbit, as a geopandas.GeoDataFrame, with columns:
          - 'velocity' : shapely.geometry.Point with velocity in x, y, z direction
          - 'geometry' : shapely.geometry.Point with position in x, y, z direction

        crs is set to 'geocentric'

        attrs keys:
          - 'orbit_pass': 'Ascending' or 'Descending'
          - 'platform_heading': in degrees, relative to north

        Notes
        -----
        orbit is longer than the SAFE, because it belongs to all datatakes, not only this slice

        """
        return self.dt["orbit"].to_dataset()

    @property
    def image(self) -> xarray.Dataset:
        return self.dt["image"].to_dataset()

    @property
    def azimuth_fmrate(self):
        """
        xarray.Dataset
            Frequency Modulation rate annotations such as t0 (azimuth time reference) and polynomial coefficients: Azimuth FM rate = c0 + c1(tSR - t0) + c2(tSR - t0)^2
        """
        return self.dt["azimuth_fmrate"].to_dataset()

    @property
    def _dict_coords2ll(self):
        """
        dict with keys ['longitude', 'latitude'] with interpolation function (RectBivariateSpline) as values.

        Examples:
        ---------
            get longitude at line=100 and sample=200:
            ```
            >>> self._dict_coords2ll["longitude"].ev(100, 200)
            array(-66.43947434)
            ```
        Notes:
        ------
            if self.cross_antemeridian is True, 'longitude' will be in range [0, 360]
        """
        resdict = {}
        geoloc = self.geoloc
        if self.cross_antemeridian:
            geoloc["longitude"] = geoloc["longitude"] % 360

        idx_sample = np.array(geoloc.sample)
        idx_line = np.array(geoloc.line)

        for ll in ["longitude", "latitude"]:
            resdict[ll] = RectBivariateSpline(
                idx_line, idx_sample, np.asarray(geoloc[ll]), kx=1, ky=1
            )

        return resdict

    @property
    def _bursts(self):
        return self.dt["bursts"].to_dataset()

    @property
    def approx_transform(self):
        """
        Affine transfom from geoloc.

        This is an inaccurate transform, with errors up to 600 meters.
        But it's fast, and may fit some needs, because the error is stable localy.
        See `xsar.BaseMeta.coords2ll` `xsar.BaseMeta.ll2coords` for accurate methods.

        Examples
        --------
            get `longitude` and `latitude` from tuple `(line, sample)`:

            >>> longitude, latitude = self.approx_transform * (line, sample)

            get `line` and `sample` from tuple `(longitude, latitude)`

            >>> line, sample = ~self.approx_transform * (longitude, latitude)

        See Also
        --------
        xsar.BaseMeta.coords2ll
        xsar.BaseMeta.ll2coords
        xsar.BaseMeta.coords2ll
        xsar.BaseMeta.ll2coords

        """
        return self.geoloc.attrs["approx_transform"]

    def __repr__(self):
        if self.multidataset:
            meta_type = "multi (%d)" % len(self.subdatasets)
        else:
            meta_type = "single"
        return "<Sentinel1Meta %s object>" % meta_type

    def _repr_mimebundle_(self, include=None, exclude=None):
        return repr_mimebundle(self, include=include, exclude=exclude)

    def __reduce__(self):
        # make self serializable with pickle
        # https://docs.python.org/3/library/pickle.html#object.__reduce__

        return self.__class__, (self.name,), self.dict

    @property
    def _doppler_estimate(self):
        """
        xarray.Dataset
            with Doppler Centroid Estimates from annotations such as geo_polynom,data_polynom or frequency
        """
        return self.dt["doppler_estimate"].to_dataset()

    @property
    def get_calibration_luts(self):
        """
        get original (ie not interpolation) xr.Dataset sigma0 and gamma0 Look Up Tables to apply calibration

        """
        return self.dt["calibration_luts"].to_dataset()

    @property
    def get_noise_azi_raw(self):
        return self.dt["noise_azimuth_raw"].to_dataset()

    @property
    def get_noise_range_raw(self):
        return self.dt["noise_range_raw"].to_dataset()

    @property
    def get_antenna_pattern(self):
        return self.dt["antenna_pattern"].to_dataset()

    @property
    def get_swath_merging(self):
        return self.dt["swath_merging"].to_dataset()
